# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin model views for records."""

import json

from flask import flash
from flask_admin.contrib.sqla import ModelView
from flask_babelex import gettext as _
from invenio_admin.filters import FilterConverter
from invenio_db import db
from markupsafe import Markup
from sqlalchemy.exc import SQLAlchemyError

from .api import Record
from .models import RecordMetadata


class RecordMetadataModelView(ModelView):
    """Records admin model view."""

    filter_converter = FilterConverter()
    can_create = False
    can_edit = False
    can_delete = True
    can_view_details = True
    column_list = ('id', 'version_id', 'updated', 'created',)
    column_details_list = ('id', 'version_id', 'updated', 'created', 'json')
    column_labels = dict(
        id=_('UUID'),
        version_id=_('Revision'),
        json=_('JSON'),
    )
    column_formatters = dict(
        version_id=lambda v, c, m, p: m.version_id-1,
        json=lambda v, c, m, p: Markup("<pre>{0}</pre>").format(
            json.dumps(m.json, indent=2, sort_keys=True))
    )
    column_filters = ('created', 'updated', )
    column_default_sort = ('updated', True)
    page_size = 25

    def delete_model(self, model):
        """Delete a record."""
        try:
            if model.json is None:
                return True
            record = Record(model.json, model=model)
            record.delete()
            db.session.commit()
        except SQLAlchemyError as e:
            if not self.handle_view_exception(e):
                flash(_('Failed to delete record. %(error)s', error=str(e)),
                      category='error')
            db.session.rollback()
            return False
        return True


record_adminview = dict(
    modelview=RecordMetadataModelView,
    model=RecordMetadata,
    category=_('Records'))




# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""InvenioRefResolver."""

import urllib.parse

from jsonschema import RefResolutionError, RefResolver


class InvenioRefResolver(RefResolver):
    """Local Invenio JSONSchemas ref resolver class."""

    def resolve_remote(self, uri):
        """Block remote ref resolve."""
        raise RefResolutionError(
            "{uri} not found in local registry.".format(uri=uri))


def urljoin_with_custom_scheme(*args, **kwargs):
    """Patch urljoin call when using local store with custom schemes.

    Allows using custom schemes by extending urllib.parse
    configuration (work-around suggested in
    https://bugs.python.org/issue18828). This patch won't
    be needed once https://github.com/Julian/jsonschema/issues/649
    gets fixed.
    """
    for arg in args:
        if isinstance(arg, str) and "://" in arg:
            scheme = arg.split("://")[0]
            if scheme not in urllib.parse.uses_relative:
                urllib.parse.uses_relative.append(scheme)
            if scheme not in urllib.parse.uses_netloc:
                urllib.parse.uses_netloc.append(scheme)

    return urllib.parse.urljoin(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field."""
from copy import deepcopy

from invenio_db import db

from ...dictutils import dict_lookup, parse_lookup_key
from .errors import InvalidRelationValue
from .results import RelationListResult, RelationNestedListResult, \
    RelationResult


class RelationBase:
    """Base class for defining relation fields."""

    result_cls = RelationResult

    def __init__(self, key=None, attrs=None, keys=None, _value_key_suffix='id',
                 _clear_empty=True, cache_key=None, value_check=None):
        """Initialize the relation."""
        self.key = key
        self.attrs = attrs or []
        self.keys = keys or []
        self._value_key_suffix = _value_key_suffix
        self._clear_empty = _clear_empty
        self._cache_key = cache_key
        self.value_check = value_check
        self._cache_ref = None

    def inject_cache(self, cache, default_key):
        """Internal method used by mapping to inject a shared cache."""
        self._cache_ref = cache
        if self._cache_key is None:
            self._cache_key = default_key
        if self._cache_key not in self._cache_ref:
            self._cache_ref[self._cache_key] = {}

    @property
    def cache(self):
        """Get the cache for this relation.

        Note the cache is shared with all relations defined on the same system
        field.
        """
        return self._cache_ref[self._cache_key]

    def resolve(self, id_):
        """Resolve a relation by its ID (has to be implemented)."""
        raise NotImplementedError()

    @property
    def value_key(self):
        """Default stored value key getter."""
        return f'{self.key}.{self._value_key_suffix}'

    def exists(self, id_):
        """Default existence check by ID."""
        return self.resolve(id_) is not None

    def exists_many(self, ids):
        """Default multiple existence check by a list of IDs."""
        return all(self.exists(i) for i in ids)

    def get_value(self, record):
        """Return the resolved relation from a record."""
        return self.result_cls(self, record)

    def parse_value(self, value):
        """Parse an object to a resolvable ID to be stored."""
        return value

    def set_value(self, record, value):
        """Set the relation value."""
        store_value = self.parse_value(value)
        if self.exists(store_value):
            # TODO: stolen from `SystemField.set_dictkey`
            keys = parse_lookup_key(self.value_key)
            try:
                parent = dict_lookup(record, keys, parent=True)
            except KeyError as e:
                parent = record
                for k in keys[:-1]:
                    if k not in parent:
                        parent[k] = {}
                    else:
                        if not isinstance(parent[k], dict):
                            raise KeyError(
                                f"Expected a dict at subkey '{k}'. "
                                f"Found '{parent[k].__class__.__name__}'."
                            )
                    parent = parent[k]

            if not isinstance(parent, dict):
                raise KeyError(
                    f"Expected a dict at subkey '{keys[-2]}'. "
                    f"Found '{parent.__class__.__name__}'."
                )

            parent[keys[-1]] = store_value
        else:
            raise InvalidRelationValue("Invalid field value.")

    def clear_value(self, record):
        """Clear the relation value."""
        keys = parse_lookup_key(self.value_key)
        try:
            parent = dict_lookup(record, keys, parent=True)
        except KeyError as e:
            parent = record
            for k in keys[:-1]:
                if k not in parent:  # Nothing to set
                    return
                parent = parent[k]
        parent.pop(keys[-1], None)
        if self._clear_empty and parent == {}:
            parent = dict_lookup(record, keys[:-1], parent=True)
            parent.pop(keys[-2], None)


class ListRelation(RelationBase):
    """Primary-key relation list type."""

    result_cls = RelationListResult

    def __init__(self, *args, relation_field=None, **kwargs):
        """Initialize the list relation.

        :param relation_field:  The field representing the relation. (Intended
                                to be used when we are passing a list of
                                objects but only one field of each object is
                                referring to a relation)
        :type relation_field: str
        """
        self.relation_field = relation_field
        super().__init__(*args, **kwargs)

    def parse_value(self, value):
        """Parse a record (or ID) to the ID to be stored."""
        if isinstance(value, (tuple, list)):
            if self.relation_field:
                return [super(ListRelation, self).parse_value(
                    v.get(self.relation_field)) for v in
                        value if self.relation_field in v]
            return [super(ListRelation, self).parse_value(v) for v in value]
        else:
            raise InvalidRelationValue('Invalid value. Expected list.')

    def _get_parent(self, record, keys):
        """Get parent dict."""
        try:
            parent = dict_lookup(record, keys[:-1], parent=True)
        except KeyError as e:
            parent = record
            for k in keys[:-2]:
                if k not in parent:
                    parent[k] = {}
                else:
                    if not isinstance(parent[k], dict):
                        raise KeyError(
                            f"Expected a dict at subkey '{k}'. "
                            f"Found '{parent[k].__class__.__name__}'."
                        )
                parent = parent[k]
        return parent

    def set_value(self, record, value):
        """Set the relation value."""
        store_values = self.parse_value(value)
        # Validate all values
        if self.exists_many(store_values):
            keys = parse_lookup_key(self.value_key)
            store_key, rel_id_key = keys[-2:]
            parent = self._get_parent(record, keys)

            if self.relation_field:
                values_list = deepcopy(value)
                for v in values_list:
                    sv = v.get(self.relation_field)
                    if sv:
                        v[self.relation_field] = {rel_id_key: sv}
            else:
                values_list = [{rel_id_key: sv} for sv in store_values]

            parent[store_key] = values_list
        else:
            raise InvalidRelationValue("Invalid values.")

    def clear_value(self, record):
        """Clear the relation value."""
        keys = parse_lookup_key(self.key)
        try:
            parent = dict_lookup(record, keys, parent=True)
        except KeyError as e:
            parent = record
            for k in keys[:-1]:
                if k not in parent:  # Nothing to set
                    return
                parent = parent[k]
        parent.pop(keys[-1], None)
        if self._clear_empty and parent == []:
            parent = dict_lookup(record, keys[:-1], parent=True)
            parent.pop(keys[-2], None)


class PKRelation(RelationBase):
    """Primary-key relation type."""

    def __init__(self, *args, record_cls=None, **kwargs):
        """Initialize the PK relation."""
        self.record_cls = record_cls
        super().__init__(*args, **kwargs)

    def resolve(self, id_):
        """Resolve the value using the record class."""
        if id_ in self.cache:
            return self.cache[id_]

        try:
            obj = self.record_cls.get_record(id_)
            # We detach the related record model from the database session when
            # we add it in the cache. Otherwise, accessing the cached record
            # model, will execute a new select query after a db.session.commit.
            db.session.expunge(obj.model)
            self.cache[id_] = obj
            return obj
        # TODO: there's many ways Record.get_record can fail...
        except Exception:
            return None

    def parse_value(self, value):
        """Parse a record (or ID) to the ID to be stored."""
        if isinstance(value, str):
            return value
        elif isinstance(value, self.record_cls):
            return str(value.id)
        else:
            raise InvalidRelationValue(
                f'Invalid value. Expected "str" or "{self.record_cls}"')

    # TODO: We could have a more efficient "exists" via PK queries
    # def exists(self, id_):
    #     """Check if an ID exists using the record class."""
    #     return bool(self.record_cls.model_cls.query.filter_by(id=id_))
    #
    # def exists_many(self, ids):
    #     """."""
    #     self.record_cls.get_record(id_)


class PKListRelation(ListRelation, PKRelation):
    """Primary-key list relation."""


class NestedListRelation(ListRelation):
    """Primary-key relation list type."""

    result_cls = RelationNestedListResult

    def exists_many(self, ids):
        """Default multiple existence check by a list of IDs."""
        return all(
            all(self.exists(i) for i in inner_ids) for inner_ids in ids
        )

    def parse_value(self, value):
        """Parse a record (or ID) to the ID to be stored."""
        if isinstance(value, (tuple, list)):
            outter_list = []
            if self.relation_field:
                for inner_list in value:
                    inner_v = inner_list.get(self.relation_field)
                    if inner_v:
                        outter_list.append([
                            super(ListRelation, self).parse_value(v)
                            for v in inner_v
                        ])
            else:
                for inner_list in value:
                    outter_list.append([
                        super(ListRelation, self).parse_value(v)
                        for v in inner_list
                    ])
            return outter_list
        else:
            raise InvalidRelationValue('Invalid value. Expected list.')

    def set_value(self, record, value):
        """Set the relation value."""
        store_values = self.parse_value(value)
        # Validate all values
        if self.exists_many(store_values):
            keys = parse_lookup_key(self.value_key)
            store_key, rel_id_key = keys[-2:]
            parent = self._get_parent(record, keys)

            if self.relation_field:
                values_list = deepcopy(value)
                for v in values_list:
                    inner_sv = v.get(self.relation_field)
                    if inner_sv:
                        v[self.relation_field] = [
                            {rel_id_key: sv} for sv in inner_sv
                        ]
            else:
                values_list = [
                    [{rel_id_key: sv} for sv in inner_values]
                    for inner_values in store_values
                ]

            parent[store_key] = values_list
        else:
            raise InvalidRelationValue("Invalid values.")


class PKNestedListRelation(NestedListRelation, PKRelation):
    """Primary-key nested list relation."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records API."""

from functools import partial

import pytest

from invenio_records.api import Record
from invenio_records.signals import after_record_delete, after_record_insert, \
    after_record_revert, after_record_update, before_record_delete, \
    before_record_insert, before_record_revert, before_record_update


@pytest.fixture()
def signals():
    """Fixtures to connect signals."""
    called = {}

    def _listener(signal_name, sender, *args, **kwargs):
        if signal_name not in called:
            called[signal_name] = 0
        called[signal_name] += 1

    after_record_delete_listener = partial(_listener, 'after_record_delete')
    after_record_insert_listener = partial(_listener, 'after_record_insert')
    after_record_revert_listener = partial(_listener, 'after_record_revert')
    after_record_update_listener = partial(_listener, 'after_record_update')
    before_record_delete_listener = partial(_listener, 'before_record_delete')
    before_record_insert_listener = partial(_listener, 'before_record_insert')
    before_record_revert_listener = partial(_listener, 'before_record_revert')
    before_record_update_listener = partial(_listener, 'before_record_update')
    before_record_insert_listener = partial(_listener, 'before_record_insert')

    after_record_delete.connect(after_record_delete_listener)
    after_record_insert.connect(after_record_insert_listener)
    after_record_revert.connect(after_record_revert_listener)
    after_record_update.connect(after_record_update_listener)
    before_record_delete.connect(before_record_delete_listener)
    before_record_insert.connect(before_record_insert_listener)
    before_record_revert.connect(before_record_revert_listener)
    before_record_update.connect(before_record_update_listener)
    before_record_insert.connect(before_record_insert_listener)

    yield called

    after_record_delete.disconnect(after_record_delete_listener)
    after_record_insert.disconnect(after_record_insert_listener)
    after_record_revert.disconnect(after_record_revert_listener)
    after_record_update.disconnect(after_record_update_listener)
    before_record_delete.disconnect(before_record_delete_listener)
    before_record_insert.disconnect(before_record_insert_listener)
    before_record_revert.disconnect(before_record_revert_listener)
    before_record_update.disconnect(before_record_update_listener)
    before_record_insert.disconnect(before_record_insert_listener)


def test_signals(testapp, database, signals):
    """Test signals being sent."""
    db = database
    record = Record.create({'title': 'Test'})
    db.session.commit()
    assert 'before_record_insert' in signals
    assert 'after_record_insert' in signals
    assert len(signals.keys()) == 2

    record['title'] = 'Test2'
    record.commit()
    db.session.commit()
    assert 'before_record_update' in signals
    assert 'after_record_update' in signals
    assert len(signals.keys()) == 4

    record.revert(0)
    db.session.commit()
    assert 'before_record_revert' in signals
    assert 'after_record_revert' in signals
    assert len(signals.keys()) == 6

    record.delete()
    db.session.commit()
    assert 'before_record_delete' in signals
    assert 'after_record_delete' in signals
    assert len(signals.keys()) == 8


def test_signals_disabled(testapp, database, signals):
    """Test signals being sent."""
    db = database

    class MyRecord(Record):
        # Disable signals
        send_signals = False

    # Same operations as above (but none of them should send signals)
    record = MyRecord.create({'title': 'Test'})
    db.session.commit()
    record['title'] = 'Test2'
    record.commit()
    db.session.commit()
    record.revert(0)
    db.session.commit()
    record.delete()
    db.session.commit()

    # Assert that no signals was sent.
    assert len(signals.keys()) == 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records."""


import uuid

import pytest
from invenio_db import InvenioDB
from invenio_db import db as db_
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.functions import create_database, database_exists

from invenio_records import Record
from invenio_records.models import RecordMetadataBase


@pytest.fixture(scope='module')
def CustomMetadata(database):
    """."""
    db = database

    class CustomMetadata(db.Model, RecordMetadataBase):
        __tablename__ = 'custom_metadata'

    CustomMetadata.__table__.create(db.engine)
    return CustomMetadata


def test_class_model(testapp, database, CustomMetadata):
    """Test custom class model."""
    db = database

    class CustomRecord(Record):
        model_cls = CustomMetadata

    assert 'custom_metadata' in db.metadata.tables.keys()

    recid = uuid.UUID('262d2748-ba41-456f-a844-4d043a419a6f')

    # Create a new record with two mutables, a list and a dict
    rec = CustomRecord.create(
        {
            'title': 'Title',
            'list': ['foo', ],
            'dict': {'moo': 'boo'},
        },
        id_=recid)

    db.session.commit()
    db.session.expunge_all()
    # record should be in the table
    rec = CustomRecord.get_record(recid)
    assert rec == {
        'title': 'Title',
        'list': ['foo', ],
        'dict': {'moo': 'boo'}
    }
    # the record should not be in the default table
    pytest.raises(NoResultFound, Record.get_record, recid)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

import pytest
from flask import Flask
from invenio_celery import InvenioCelery
from invenio_db import InvenioDB
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DropConstraint, DropSequence, DropTable

from invenio_records import InvenioRecords

pytest_plugins = ("celery.contrib.pytest", )


@compiles(DropTable, 'postgresql')
def _compile_drop_table(element, compiler, **kwargs):
    return compiler.visit_drop_table(element) + ' CASCADE'


@compiles(DropConstraint, 'postgresql')
def _compile_drop_constraint(element, compiler, **kwargs):
    return compiler.visit_drop_constraint(element) + ' CASCADE'


@compiles(DropSequence, 'postgresql')
def _compile_drop_sequence(element, compiler, **kwargs):
    return compiler.visit_drop_sequence(element) + ' CASCADE'


@pytest.fixture(scope='module')
def create_app(instance_path):
    """Application factory fixture for use with pytest-invenio."""
    def _create_app(**config):
        app_ = Flask(
            __name__,
            instance_path=instance_path,
        )
        app_.config.update(config)
        InvenioCelery(app_)
        InvenioDB(app_)
        InvenioRecords(app_)
        return app_
    return _create_app


@pytest.fixture(scope='module')
def testapp(base_app, database):
    """Application with just a database.

    Pytest-Invenio also initialises ES with the app fixture.
    """
    yield base_app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations dumper.

Dumper used to dump/load relations to/from an ElasticSearch body.
"""

from ..dictutils import dict_lookup, dict_set
from .elasticsearch import ElasticsearchDumperExt


class RelationDumperExt(ElasticsearchDumperExt):
    """Dumper for a relations field."""

    def __init__(self, key, fields=None):
        """Initialize the dumper."""
        self.key = key
        self.fields = fields

    def dump(self, record, data):
        """Dump relations."""
        relations = getattr(record, self.key)
        relations.dereference(fields=self.fields)
        relation_fields = self.fields or relations
        for rel_field_name in relation_fields:
            rel_field = getattr(relations, rel_field_name)
            try:
                dict_set(
                    data, rel_field.key,
                    dict_lookup(record, rel_field.key)
                )
            except KeyError:
                pass

    def load(self, data, record_cls):
        """Load relations."""
        # TODO: figure out what/how to load
        pass





# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test for model field.

This tests needs to live in it's own file to have a clean database session.
"""

import pytest
from sqlalchemy_utils.types import UUIDType

from invenio_records.api import Record
from invenio_records.dumpers import ElasticsearchDumper
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.models import RecordMetadataBase
from invenio_records.systemfields import ModelField, ModelRelation, \
    RelationsField, SystemFieldsMixin
from invenio_records.systemfields.relations.modelrelations import \
    InvalidRelationValue, ModelRelation


@pytest.fixture(scope='module')
def Record3Metadata(database):
    """."""
    db = database

    class Record3Metadata(db.Model, RecordMetadataBase):
        __tablename__ = 'record3_metadata'

    Record3Metadata.__table__.create(db.engine)
    return Record3Metadata


@pytest.fixture(scope='module')
def Record2Metadata(Record3Metadata, database):
    """."""
    db = database

    class Record2Metadata(db.Model, RecordMetadataBase):
        __tablename__ = 'record2_metadata'

        record3_id = db.Column(UUIDType, db.ForeignKey(Record3Metadata.id))

    Record2Metadata.__table__.create(db.engine)
    return Record2Metadata


def test_model_relation(testapp, database, Record3Metadata, Record2Metadata):
    """Test model field with clearing none/empty values."""
    db = database

    class Record3(Record, SystemFieldsMixin):
        model_cls = Record3Metadata
        dumper = ElasticsearchDumper()

    class Record2(Record, SystemFieldsMixin):
        model_cls = Record2Metadata
        dumper = ElasticsearchDumper(extensions=[
            RelationDumperExt('relations'),
        ])
        record3_id = ModelField('record3_id')
        relations = RelationsField(
            record3=ModelRelation(
                Record3,
                'record3_id',
                'record3',
                keys=['title'],
            )
        )

    # Set relation via model field
    rec3 = Record3.create({'title': 'this is record 3', 'description': 'n/a'})
    rec3.commit()

    rec2 = Record2.create({'title': 'this is record 2'})
    rec2.record3_id = rec3.id
    rec2.relations.dereference()
    rec2.commit()

    # We do not store the dereferenced data in the record.
    assert 'record3' not in rec2

    # Dumping dereferences the data:
    dump = rec2.dumps()
    assert dump['record3_id'] == str(rec3.id)
    assert dump['record3'] == {
            'id': str(rec3.id),
            'title': 'this is record 3',
            '@v': f'{rec3.id}::{rec3.revision_id}',
    }

    rec2 = Record2.create({'title': 'this is record 2'})
    rec2.relations.dereference()
    rec2.commit()
    assert rec2.record3_id is None

    # Set relation via relations field (record and model class)
    rec2 = Record2.create({'title': 'this is record 2'})
    rec2.relations.record3 = rec3
    rec2.relations.record3 = rec3.model
    rec2.commit()

    # Set and invalid object
    assert pytest.raises(
        InvalidRelationValue,
        setattr,
        rec2.relations,
        'record3',
        'someid',
    )

    # Unset a relation
    rec2 = Record2.create({'title': 'this is record 2'})
    rec2.relations.record3 = rec3
    rec2.commit()
    del rec2.relations.record3
    assert rec2.record3_id is None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test the dumpers API."""

from datetime import date, datetime
from uuid import UUID

import pytest
from sqlalchemy.dialects import mysql

from invenio_records.api import Record
from invenio_records.dumpers import ElasticsearchDumper, ElasticsearchDumperExt
from invenio_records.dumpers.indexedat import IndexedAtDumperExt
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.models import RecordMetadataBase
from invenio_records.systemfields.relations import PKListRelation, \
    PKRelation, RelationsField


@pytest.fixture()
def example_data():
    """Example record used for tests."""
    return {
        # "$schema": "",
        "id": "12345-abcde",
        "metadata": {
            "title": "My record",
            "date": "2020-09-20",
        },
        "pids": {
            "oaiid": {"value": "", "provider": "local"},
        },
    }


@pytest.fixture()
def es_hit():
    """Example record used for tests."""
    return {
        "_index": "testindex",
        "_type": "_doc",
        "_id": "4beb3b3e-a935-442e-a47b-6d386947ea20",
        "_version": 5,
        "_seq_no": 0,
        "_primary_term": 1,
        "found": True,
        "_source": {
            "uuid": "4beb3b3e-a935-442e-a47b-6d386947ea20",
            "version_id": 4,
            "created": "2020-09-01T14:26:00+00:00",
            "updated": "2020-09-02T14:28:21.968149+00:00'",
            "id": "12345-abcde",
            "metadata": {
                "title": "My record",
                "date": "2020-09-20",
            },
            "pids": {
                "oaiid": {"value": "", "provider": "local"},
            },
        }
    }


def test_esdumper_without_model(testapp, db, example_data):
    """Test the Elasticsearch dumper."""
    # Dump without a model.
    dump = Record(example_data).dumps(dumper=ElasticsearchDumper())
    for k in ['uuid', 'version_id', 'created', 'updated']:
        assert dump[k] is None  # keys is set to none without a model
    # Load without a model defined
    record = Record.loads(dump, loader=ElasticsearchDumper())
    assert record.model is None  # model will not be set
    assert record == example_data  # data is equivalent to initial data


def test_esdumper_with_model(testapp, db, example_data):
    """Test the Elasticsearch dumper."""
    # Create a record
    record = Record.create(example_data)
    db.session.commit()

    # Dump it
    dump = record.dumps(dumper=ElasticsearchDumper())
    assert dump['uuid'] == str(record.id)
    assert dump['version_id'] == record.revision_id + 1
    assert dump['created'][:19] == record.created.isoformat()[:19]
    assert dump['updated'][:19] == record.updated.isoformat()[:19]

    # Load it
    new_record = Record.loads(dump, loader=ElasticsearchDumper())
    assert new_record == record
    assert new_record.id == record.id
    assert new_record.revision_id == record.revision_id
    assert new_record.created == record.created
    assert new_record.updated == record.updated
    assert new_record.model.json == record.model.json


def test_esdumper_with_extensions(testapp, db, example_data):
    """Test extensions implementation."""
    # Create a simple extension that adds a computed field.
    class TestExt(ElasticsearchDumperExt):
        def dump(self, record, data):
            data['count'] = len(data['mylist'])

        def load(self, data, record_cls):
            data.pop('count')

    dumper = ElasticsearchDumper(extensions=[TestExt()])

    # Create the record
    record = Record.create({'mylist': ['a', 'b']})
    db.session.commit()

    # Dump it
    dump = record.dumps(dumper=dumper)
    assert dump['count'] == 2

    # Load it
    new_record = Record.loads(dump, loader=dumper)
    assert 'count' not in new_record


def test_esdumper_sa_datatypes(testapp, database):
    """Test to determine the data type of an SQLAlchemy field."""
    db = database

    class Model(db.Model, RecordMetadataBase):
        string = db.Column(db.String(255))
        text = db.Column(db.Text)
        biginteger = db.Column(db.BigInteger)
        integer = db.Column(db.Integer)
        boolean = db.Column(db.Boolean(name='boolean'))
        text_variant = db.Column(db.Text().with_variant(
            mysql.VARCHAR(255), 'mysql'))

    assert ElasticsearchDumper._sa_type(Model, 'biginteger') == int
    assert ElasticsearchDumper._sa_type(Model, 'boolean') == bool
    assert ElasticsearchDumper._sa_type(Model, 'created') == datetime
    assert ElasticsearchDumper._sa_type(Model, 'id') == UUID
    assert ElasticsearchDumper._sa_type(Model, 'integer') == int
    assert ElasticsearchDumper._sa_type(Model, 'json') == dict
    assert ElasticsearchDumper._sa_type(Model, 'text_variant') == str
    assert ElasticsearchDumper._sa_type(Model, 'text') == str
    assert ElasticsearchDumper._sa_type(Model, 'updated') == datetime
    assert ElasticsearchDumper._sa_type(Model, 'invalid') is None


def test_relations_dumper(testapp, db, example_data):
    """Test relations dumper extension."""
    class RecordWithRelations(Record):
        relations = RelationsField(
            language=PKRelation(
                key='language', keys=['iso', 'information.ethnicity'],
                record_cls=Record),
            languages=PKListRelation(
                key='languages', keys=['iso', 'information.ethnicity'],
                record_cls=Record),
        )

        dumper = ElasticsearchDumper(
            extensions=[RelationDumperExt('relations')])

    # Create the record
    en_language = Record.create(
        {'title': 'English', 'iso': 'en', 'information': {
            'native_speakers': '400 million', 'ethnicity': 'English'}})
    fr_language = Record.create(
        {'title': 'French', 'iso': 'fr', 'information': {
            'native_speakers': '76.8 million', 'ethnicity': 'French'}})
    db.session.commit()
    record = RecordWithRelations.create({
        'foo': 'bar',
        'mylist': ['a', 'b'],
    })
    record.relations.language = en_language
    record.relations.languages = [en_language, fr_language]
    db.session.commit()

    # Dump it
    dump = record.dumps()
    assert dump['foo'] == 'bar'
    assert dump['mylist'] == ['a', 'b']
    assert dump['language'] == {
        'id': str(en_language.id),
        'iso': 'en',
        'information': {'ethnicity': 'English'},
        '@v': str(en_language.id) + '::' + str(en_language.revision_id)
    }
    assert dump['languages'] == [
        {
            'id': str(en_language.id),
            'iso': 'en',
            'information': {'ethnicity': 'English'},
            '@v': str(en_language.id) + '::' + str(en_language.revision_id)
        },
        {
            'id': str(fr_language.id),
            'iso': 'fr',
            'information': {'ethnicity': 'French'},
            '@v': str(fr_language.id) + '::' + str(fr_language.revision_id)
        },
    ]
    assert dump['uuid'] == str(record.id)
    assert dump['version_id'] == record.revision_id + 1
    assert dump['created'][:19] == record.created.isoformat()[:19]
    assert dump['updated'][:19] == record.updated.isoformat()[:19]

    # TODO: Implement loader
    # Load it
    # new_record = Record.loads(dump, loader=dumper)
    # assert 'count' not in new_record


def test_indexedtime_dumper(testapp, db, example_data):
    """Test relations dumper extension."""
    class RecordWithIndexedTime(Record):

        dumper = ElasticsearchDumper(
            extensions=[IndexedAtDumperExt()]
        )

    # create the record
    record = RecordWithIndexedTime.create({
        'foo': 'bar',
        'mylist': ['a', 'b'],
    })
    db.session.commit()

    # dump it
    dump = record.dumps()
    assert dump['foo'] == 'bar'
    assert dump['mylist'] == ['a', 'b']
    assert dump['uuid'] == str(record.id)
    assert dump['version_id'] == record.revision_id + 1
    assert dump['created'][:19] == record.created.isoformat()[:19]
    assert dump['updated'][:19] == record.updated.isoformat()[:19]
    # only feasible to check the date (not time)
    assert dump['indexed_at'][:19] == record.created.isoformat()[:19]

    # load it
    new_record = RecordWithIndexedTime.loads(dump)
    assert 'indexed_at' not in new_record



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
# Copyright (C) 2021 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record API."""


import inspect
import warnings
from copy import deepcopy

from flask import current_app
from invenio_db import db
from jsonpatch import apply_patch
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_continuum.utils import parent_class
from werkzeug.local import LocalProxy

from .dictutils import clear_none, dict_lookup
from .dumpers import Dumper
from .errors import MissingModelError
from .models import RecordMetadata
from .signals import after_record_delete, after_record_insert, \
    after_record_revert, after_record_update, before_record_delete, \
    before_record_insert, before_record_revert, before_record_update

_records_state = LocalProxy(lambda: current_app.extensions['invenio-records'])


class RecordBase(dict):
    """Base class for Record and RecordRevision to share common features."""

    model_cls = RecordMetadata
    """SQLAlchemy model class defining which table stores the records."""

    format_checker = None
    """Class-level attribute to specify a default JSONSchema format checker."""

    validator = None
    """Class-level attribute to specify a JSONSchema validator class."""

    dumper = Dumper()
    """Class-level attribute to specify the default data dumper/loader.

    For backward compatibility the dumper used here just produces a deep copy
    of the record.
    """

    enable_jsonref = True
    """Class-level attribute to control if JSONRef replacement is supported."""

    _extensions = []
    """Record extensions registry.

    Allows extensions (like system fields) to be registered on the record.
    """

    def __init__(self, data, model=None, **kwargs):
        """Initialize instance with dictionary data and SQLAlchemy model.

        :param data: Dict with record metadata.
        :param model: :class:`~invenio_records.models.RecordMetadata` instance.
        """
        self.model = model
        for e in self._extensions:
            e.pre_init(self, data, model=model, **kwargs)
        super(RecordBase, self).__init__(data or {})
        for e in self._extensions:
            e.post_init(self, data, model=model, **kwargs)

    @property
    def id(self):
        """Get model identifier."""
        return self.model.id if self.model else None

    @property
    def revision_id(self):
        """Get revision identifier."""
        return self.model.version_id-1 if self.model else None

    @property
    def created(self):
        """Get creation timestamp."""
        return self.model.created if self.model else None

    @property
    def updated(self):
        """Get last updated timestamp."""
        return self.model.updated if self.model else None

    @property
    def is_deleted(self):
        """Get creation timestamp."""
        return self.model.is_deleted if self.model else None

    def validate(self, format_checker=None, validator=None, **kwargs):
        r"""Validate record according to schema defined in ``$schema`` key.

        :Keyword Arguments:
          * **format_checker** --
            A ``format_checker`` is an instance of class
            :class:`jsonschema.FormatChecker` containing business logic to
            validate arbitrary formats. For example:

            >>> from jsonschema import FormatChecker
            >>> from jsonschema.validators import validate
            >>> checker = FormatChecker()
            >>> checker.checks('foo')(lambda el: el.startswith('foo'))
            <function <lambda> at ...>
            >>> validate('foo', {'format': 'foo'}, format_checker=checker)

            returns ``None``, which means that the validation was successful,
            while

            >>> validate('bar', {'format': 'foo'},
            ...    format_checker=checker)  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
              ...
            ValidationError: 'bar' is not a 'foo'
              ...

            raises a :class:`jsonschema.exceptions.ValidationError`.

          * **validator** --
            A :class:`jsonschema.protocols.Validator` class used for record
            validation. It will be used as `cls` argument when calling
            :func:`jsonschema.validate`. For example

            >>> from jsonschema.validators import extend, Draft4Validator
            >>> NoRequiredValidator = extend(
            ...     Draft4Validator,
            ...     validators={'required': lambda v, r, i, s: None}
            ... )
            >>> schema = {
            ...     'type': 'object',
            ...     'properties': {
            ...         'name': { 'type': 'string' },
            ...         'email': { 'type': 'string' },
            ...         'address': {'type': 'string' },
            ...         'telephone': { 'type': 'string' }
            ...     },
            ...     'required': ['name', 'email']
            ... }
            >>> from jsonschema.validators import validate
            >>> validate({}, schema, NoRequiredValidator)

            returns ``None``, which means that the validation was successful,
            while

            >>> validate({}, schema)  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            ValidationError: 'name' is a required property
            ...

            raises a :class:`jsonschema.exceptions.ValidationError`.
        """
        # 1) For backward compatibility we do not change the method signature
        # (i.e. return a ``None`` value on successful validation).
        # The actual implementation of the validation method is implemented
        # below in _validate() which is also the one used internally to avoid
        # double encoding of the dict to JSON.
        # 2) We ignore **kwargs (but keep it for backward compatibility) as
        # the jsonschema.protocols.Validator only takes the two keyword
        # arguments formater_checker and cls (i.e. validator).
        self._validate(format_checker=format_checker, validator=validator)

    def _validate(self, format_checker=None, validator=None, use_model=False):
        """Implementation of the JSONSchema validation."""
        # Use the encoder to transform Python dictionary into JSON document
        # prior to validation unless we explicitly ask to use the already
        # encoded JSON in the model.
        if use_model:
            json = self.model.json
        else:
            json = self.model_cls.encode(dict(self))

        if '$schema' in self and self['$schema'] is not None:
            # Validate (an error will raise an exception)
            _records_state.validate(
                json,
                self['$schema'],
                # Use defaults of class if not specified by user.
                format_checker=format_checker or self.format_checker,
                cls=validator or self.validator
            )

        # Return encoded data, so we don't have to double encode.
        return json

    def replace_refs(self):
        """Replace the ``$ref`` keys within the JSON."""
        if self.enable_jsonref:
            return _records_state.replace_refs(self)
        else:
            return self

    def clear_none(self, key=None):
        """Helper method to clear None, empty dict and list values.

        Modifications are done in place.
        """
        clear_none(dict_lookup(self, key) if key else self)

    def dumps(self, dumper=None):
        """Make a dump of the record (defaults to a deep copy of the dict).

        This method produces a version of a record that can be persisted on
        storage such as the database, Elasticsearch or other mediums depending
        on the dumper class used.

        :param dumper: Dumper to use when dumping the record.
        :returns: A ``dict``.
        """
        dumper = dumper or self.dumper

        data = {}

        # Run pre dump extensions
        for e in self._extensions:
            pre_dump_params = inspect.signature(e.pre_dump).parameters
            if 'data' in pre_dump_params:
                e.pre_dump(self, data, dumper=dumper)
            else:
                # TODO: Remove in v1.6.0 or later
                warnings.warn(
                    "The pre_dump hook must take a positional argument data.",
                    DeprecationWarning
                )
                e.pre_dump(self, dumper=dumper)

        dump_params = inspect.signature(dumper.dump).parameters
        if 'data' in dump_params:
            # Execute the dump - for backwards compatibility we use the default
            # dumper which returns a deepcopy.
            data = dumper.dump(self, data)
        else:
            # TODO: Remove in v1.6.0 or later
            warnings.warn(
                "The dumper.dump() must take a positional argument data.",
                DeprecationWarning
            )
            data = dumper.dump(self)

        for e in self._extensions:
            e.post_dump(self, data, dumper=dumper)

        return data

    @classmethod
    def loads(cls, data, loader=None):
        """Load a record dump.

        :param loader: Loader class to use when loading the record.
        :returns: A new :class:`Record` instance.
        """
        # The method is named with in plural to align with dumps (which is
        # named with s even if it should probably have been called "dump"
        # instead.
        loader = loader or cls.dumper

        data = deepcopy(data)  # avoid mutating the original object
        # Run pre load extensions
        for e in cls._extensions:
            e.pre_load(data, loader=loader)

        record = loader.load(data, cls)

        # Run post load extensions
        for e in cls._extensions:
            post_load_params = inspect.signature(e.post_load).parameters
            if 'data' in post_load_params:
                e.post_load(record, data, loader=loader)
            else:
                # TODO: Remove in v1.6.0 or later
                warnings.warn(
                    "The post_load hook must take a positional argument data.",
                    DeprecationWarning
                )
                e.post_load(record, loader=loader)

        return record


class Record(RecordBase):
    """Define API for metadata creation and manipulation."""

    send_signals = True
    """Class-level attribute to control if signals should be sent."""

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        r"""Create a new record instance and store it in the database.

        #. Send a signal :data:`invenio_records.signals.before_record_insert`
           with the new record as parameter.

        #. Validate the new record data.

        #. Add the new record in the database.

        #. Send a signal :data:`invenio_records.signals.after_record_insert`
           with the new created record as parameter.

        :Keyword Arguments:
          * **format_checker** --
            An instance of the class :class:`jsonschema.FormatChecker`, which
            contains validation rules for formats. See
            :func:`~invenio_records.api.RecordBase.validate` for more details.

          * **validator** --
            A :class:`jsonschema.protocols.Validator` class that will be used
            to validate the record. See
            :func:`~invenio_records.api.RecordBase.validate` for more details.

        :param data: Dict with the record metadata.
        :param id_: Specify a UUID to use for the new record, instead of
                    automatically generated.
        :returns: A new :class:`Record` instance.
        """
        with db.session.begin_nested():
            # For backward compatibility we pop them here.
            format_checker = kwargs.pop('format_checker', None)
            validator = kwargs.pop('validator', None)

            # Create the record and the model
            record = cls(
                data,
                model=cls.model_cls(id=id_, data=data),
                **kwargs
            )

            if cls.send_signals:
                before_record_insert.send(
                    current_app._get_current_object(),
                    record=record
                )

            # Run pre create extensions
            for e in cls._extensions:
                e.pre_create(record)

            # Validate also encodes the data
            record._validate(
                format_checker=format_checker,
                validator=validator,
                use_model=True  # use model (already encoded) and didn't change
            )

            db.session.add(record.model)

        if cls.send_signals:
            after_record_insert.send(
                current_app._get_current_object(),
                record=record
            )

        # Run post create extensions
        for e in cls._extensions:
            e.post_create(record)

        return record

    @classmethod
    def get_record(cls, id_, with_deleted=False):
        """Retrieve the record by id.

        Raise a database exception if the record does not exist.

        :param id_: record ID.
        :param with_deleted: If `True` then it includes deleted records.
        :returns: The :class:`Record` instance.
        """
        with db.session.no_autoflush:
            query = cls.model_cls.query.filter_by(id=id_)
            if not with_deleted:
                query = query.filter(cls.model_cls.is_deleted != True)  # noqa
            obj = query.one()
            return cls(obj.data, model=obj)

    @classmethod
    def get_records(cls, ids, with_deleted=False):
        """Retrieve multiple records by id.

        :param ids: List of record IDs.
        :param with_deleted: If `True` then it includes deleted records.
        :returns: A list of :class:`Record` instances.
        """
        with db.session.no_autoflush:
            query = cls.model_cls.query.filter(cls.model_cls.id.in_(ids))
            if not with_deleted:
                query = query.filter(cls.model_cls.is_deleted != True)  # noqa

            return [cls(obj.data, model=obj) for obj in query.all()]

    def patch(self, patch):
        """Patch record metadata.

        :params patch: Dictionary of record metadata.
        :returns: A new :class:`Record` instance.
        """
        warnings.warn(
            "The patch() method is deprecated and will be removed.",
            DeprecationWarning
        )
        data = apply_patch(dict(self), patch)
        return self.__class__(data, model=self.model)

    def commit(self, format_checker=None, validator=None, **kwargs):
        r"""Store changes of the current record instance in the database.

        #. Send a signal :data:`invenio_records.signals.before_record_update`
           with the current record to be committed as parameter.

        #. Validate the current record data.

        #. Commit the current record in the database.

        #. Send a signal :data:`invenio_records.signals.after_record_update`
            with the committed record as parameter.

        :Keyword Arguments:
          * **format_checker** --
            An instance of the class :class:`jsonschema.FormatChecker`, which
            contains validation rules for formats. See
            :func:`~invenio_records.api.RecordBase.validate` for more details.

          * **validator** --
            A :class:`jsonschema.protocols.Validator` class that will be used
            to validate the record. See
            :func:`~invenio_records.api.RecordBase.validate` for more details.

        :returns: The :class:`Record` instance.
        """
        if self.model is None or self.model.is_deleted:
            raise MissingModelError()

        with db.session.begin_nested():
            if self.send_signals:
                before_record_update.send(
                    current_app._get_current_object(),
                    record=self
                )

            # Run pre commit extensions
            for e in self._extensions:
                e.pre_commit(self, **kwargs)

            # Validate also encodes the data
            json = self._validate(
                format_checker=format_checker, validator=validator)

            # Thus, we pass the encoded JSON directly to the model to avoid
            # double encoding.
            self.model.json = json
            flag_modified(self.model, 'json')

            db.session.merge(self.model)

        if self.send_signals:
            after_record_update.send(
                current_app._get_current_object(),
                record=self
            )

        # Run post commit extensions
        for e in self._extensions:
            e.post_commit(self)

        return self

    def delete(self, force=False):
        """Delete a record.

        If `force` is ``False``, the record is soft-deleted: record data will
        be deleted but the record identifier and the history of the record will
        be kept. This ensures that the same record identifier cannot be used
        twice, and that you can still retrieve its history. If `force` is
        ``True``, then the record is completely deleted from the database.

        #. Send a signal :data:`invenio_records.signals.before_record_delete`
           with the current record as parameter.

        #. Delete or soft-delete the current record.

        #. Send a signal :data:`invenio_records.signals.after_record_delete`
           with the current deleted record as parameter.

        :param force: if ``True``, completely deletes the current record from
               the database, otherwise soft-deletes it.
        :returns: The deleted :class:`Record` instance.
        """
        if self.model is None:
            raise MissingModelError()

        with db.session.begin_nested():
            if self.send_signals:
                before_record_delete.send(
                    current_app._get_current_object(),
                    record=self
                )

            # Run pre delete extensions
            for e in self._extensions:
                e.pre_delete(self, force=force)

            if force:
                db.session.delete(self.model)
            else:
                self.model.is_deleted = True
                db.session.merge(self.model)

        if self.send_signals:
            after_record_delete.send(
                current_app._get_current_object(),
                record=self
            )

        # Run post delete extensions
        for e in self._extensions:
            e.post_delete(self, force=force)

        return self

    def undelete(self):
        """Undelete a soft-deleted record."""
        if self.model is None:
            raise MissingModelError()

        self.model.is_deleted = False

        return self

    def revert(self, revision_id):
        """Revert the record to a specific revision.

        #. Send a signal :data:`invenio_records.signals.before_record_revert`
           with the current record as parameter.

        #. Revert the record to the revision id passed as parameter.

        #. Send a signal :data:`invenio_records.signals.after_record_revert`
           with the reverted record as parameter.

        :param revision_id: Specify the record revision id
        :returns: The :class:`Record` instance corresponding to the revision id
        """
        if self.model is None:
            raise MissingModelError()

        revision = self.revisions[revision_id]

        with db.session.begin_nested():
            if self.send_signals:
                # TODO: arguments to this signal does not make sense.
                # Ought to be both record and revision.
                before_record_revert.send(
                    current_app._get_current_object(),
                    record=self
                )

            for e in self._extensions:
                e.pre_revert(self, revision)

            # Here we explicitly set the json column in order to not
            # encode/decode the json data via the ``data`` property.
            self.model.json = revision.model.json
            flag_modified(self.model, 'json')

            db.session.merge(self.model)

        if self.send_signals:
            # TODO: arguments to this signal does not make sense.
            # Ought to be the class being returned just below and should
            # include the revision.
            after_record_revert.send(
                current_app._get_current_object(),
                record=self
            )

        record = self.__class__(self.model.data, model=self.model)

        for e in self._extensions:
            e.post_revert(record, revision)

        return record

    @property
    def revisions(self):
        """Get revisions iterator."""
        if self.model is None:
            raise MissingModelError()

        return RevisionsIterator(self.model)


class RecordRevision(RecordBase):
    """API for record revisions."""

    def __init__(self, model):
        """Initialize instance with the SQLAlchemy model."""
        super(RecordRevision, self).__init__(
            # The version model class does not have the properties of the
            # parent model class, and thus ``model.data`` won't work (which is
            # a Python property on RecordMetadataBase).
            parent_class(model.__class__).decode(model.json),
            model=model
        )


class RevisionsIterator(object):
    """Iterator for record revisions."""

    def __init__(self, model):
        """Initialize instance with the SQLAlchemy model."""
        self._it = None
        self.model = model

    def __len__(self):
        """Get number of revisions."""
        return self.model.versions.count()

    def __iter__(self):
        """Get iterator."""
        self._it = iter(self.model.versions)
        return self

    def __next__(self):
        """Get next revision item."""
        return RecordRevision(next(self._it))

    def __getitem__(self, revision_id):
        """Get a specific revision.

        Revision id is always smaller by 1 from version_id. This was initially
        to ensure that record revisions was zero-indexed similar to arrays
        (e.g. you could do ``record.revisions[0]``). Due to SQLAlchemy
        increasing the version counter via Python instead of the SQL
        insert/update query it's possible to have an "array with holes" and
        thus having it zero-indexed does not make much sense (thus it's like
        this for historical reasons and has not been changed because it's
        diffcult to change - e.g. implies all indexed records in existing
        instances having to be updated.)
        """
        if revision_id < 0:
            return RecordRevision(self.model.versions[revision_id])
        try:
            return RecordRevision(
                self.model.versions.filter_by(
                    version_id=revision_id + 1
                ).one()
            )
        except NoResultFound:
            raise IndexError

    def __contains__(self, revision_id):
        """Test if revision exists."""
        try:
            self[revision_id]
            return True
        except IndexError:
            return False

    def __reversed__(self):
        """Allows to use reversed operator."""
        for version_index in range(self.model.versions.count()):
            yield RecordRevision(self.model.versions[-(version_index+1)])




# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Test JSON resolver."""


import jsonresolver


@jsonresolver.route('/<item>', host='nest.ed')
def test_resolver(item):
    """Create a nested JSON."""
    next_ = {'$ref': 'http://nest.ed/{}'.format(item[1:]), } if len(item[1:]) \
        else '.'
    return {'letter': item[0], 'next': next_}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Elasticsearch source dumper.

Dumper used to dump/load an the body of an Elasticsearch document.
"""

from copy import deepcopy
from datetime import date, datetime
from uuid import UUID

import arrow
import pytz
from sqlalchemy.sql.sqltypes import JSON, Boolean, DateTime, Integer, String, \
    Text
from sqlalchemy.sql.type_api import Variant
from sqlalchemy_utils.types.uuid import UUIDType

from ..systemfields.model import ModelField
from .base import Dumper


class ElasticsearchDumperExt:
    """Interface for Elasticsearch dumper extensions."""

    def dump(self, record, data):
        """Dump the data."""

    def load(self, data, record_cls):
        """Load the data.

        Reverse the changes made by the dump method.
        """


class ElasticsearchDumper(Dumper):
    """Elasticsearch source dumper."""

    def __init__(self, extensions=None, model_fields=None):
        """."""
        self._extensions = extensions or []
        self._model_fields = {
            'id': ('uuid', UUID),
            'version_id': ('version_id', int),
            'created': ('created', datetime),
            'updated': ('updated', datetime),
            # is_deleted is purposely not added (deleted record isnt indexed)
        }
        self._model_fields.update(model_fields or {})

    @staticmethod
    def _sa_type(model_cls, model_field_name):
        """Introspection of SQLAlchemy column data type.

        :param model_cls: The SQLALchemy model.
        :param model_field_name: The name of the field on the SQLAlchemy model.
        """
        try:
            sa_type = \
                model_cls.__table__.columns[model_field_name].type
            sa_type_class = sa_type.__class__

            # Deal with variant class
            if issubclass(sa_type_class, Variant):
                sa_type = sa_type.impl
                sa_type_class = sa_type.__class__

            if issubclass(sa_type_class, DateTime):
                return datetime
            elif issubclass(sa_type_class, Boolean):
                return bool
            elif issubclass(sa_type_class, Integer):
                return int
            elif issubclass(sa_type_class, UUIDType):
                return UUID
            elif issubclass(sa_type_class, String):
                return str
            elif issubclass(sa_type_class, JSON):
                return dict
            return None
        except (KeyError, AttributeError):
            return None

    @staticmethod
    def _serialize(value, dump_type):
        """Serialize a value according to it's data type.

        :param value: Value to serialize.
        :param dump_type: Data type use for serialization (supported: str, int,
            bool, float, datetime, date, uuid).
        """
        if value is None:
            return value
        if dump_type in (datetime, ):
            return pytz.utc.localize(value).isoformat()
        elif dump_type in (UUID, ):
            return str(value)
        elif dump_type is not None:
            return dump_type(value)
        return value

    @staticmethod
    def _deserialize(value, dump_type):
        """Deserialize a value according to it's data type.

        :param value: Value to deserialize.
        :param dump_type: Data type use for deserialization (supported: str,
            int, bool, float, datetime, date, uuid).
        """
        if value is None:
            return value
        if dump_type in (datetime, ):
            return arrow.get(value).datetime.replace(tzinfo=None)
        elif dump_type in (UUID, ):
            return dump_type(value)
        elif dump_type is not None:
            return dump_type(value)
        return value

    def _dump_model_field(self, record, model_field_name, dump, dump_key,
                          dump_type):
        """Helper method to dump model fields.

        :param record: The record being dumped.
        :param model_field_name: The name of the SQLAlchemy model field on the
            record's model.
        :param dump: The dictionary of the current dump.
        :param dump_key: The key to use in the dump.
        :param dump_type: The data type used for serialization.
        """
        # If model is not defined, we dump None into the field value.
        if record.model is None:
            dump[dump_key] = None
            return

        # Retrieve value of the field on the model.
        val = getattr(record.model, model_field_name)

        # Determine data type if not set.
        if dump_type is None:
            dump_type = self._sa_type(record.model_cls, model_field_name)

        # Serialize (according to data type) and set value in output on the
        # specified key.
        dump[dump_key] = self._serialize(val, dump_type)

    def _load_model_field(self, record_cls, model_field_name, dump, dump_key,
                          dump_type):
        """Helper method to load model fields from dump.

        :param record_cls: The record class being used for loading.
        :param model_field_name: The name of the SQLAlchemy model field on the
            record's model.
        :param dump: The dictionary of the dump.
        :param dump_key: The key to use in the dump.
        :param dump_type: The data type used for deserialization.
        """
        # Retrieve the value
        val = dump.pop(dump_key)

        # Return None values immediately.
        if val is None:
            return val

        # Determine dump data type if not provided
        if dump_type is None:
            sa_field = getattr(record_cls.model_cls, model_field_name)
            dump_type = self._sa_type(record_cls.model_cls, model_field_name)

        # Deserialize the value
        return self._deserialize(val, dump_type)

    @staticmethod
    def _iter_modelfields(record_cls):
        """Internal helper method to extract all model fields."""
        for attr_name in dir(record_cls):
            systemfield = getattr(record_cls, attr_name)
            if isinstance(systemfield, ModelField):
                if systemfield.dump:
                    yield systemfield

    def dump(self, record, data):
        """Dump a record.

        The method adds the following keys (if the record has an associated
        model):

        - ``uuid`` - UUID of the record.
        - ``version_id`` -  the revision id of the record.
        - ``created`` - Creation timestamp in UTC.
        - ``updated`` - Modification timestamp in UTC.
        """
        # Copy data first, otherwise we modify the record.
        dump_data = super().dump(record, data)

        # Dump model fields explicitly requested
        it = self._model_fields.items()
        for model_field_name, (dump_key, dump_type) in it:
            self._dump_model_field(
                record,
                model_field_name,
                dump_data,
                dump_key,
                dump_type,
            )

        # Dump model fields defined as system fields.
        for systemfield in self._iter_modelfields(record.__class__):
            self._dump_model_field(
                record,
                systemfield.model_field_name,
                dump_data,
                systemfield.dump_key,
                systemfield.dump_type,
            )

        # Allow extensions to integrate as well.
        for e in self._extensions:
            e.dump(record, dump_data)

        return dump_data

    def load(self, dump_data, record_cls):
        """Load a record from an Elasticsearch document source.

        The method reverses the changes made during the dump. If a model was
        associated, a model will also be initialized.

        .. warning::
            The model is not added to the SQLAlchemy session. If you plan on
            using the model, you must merge it into the session using e.g.:

            .. code-block:: python

                db.session.merge(record.model)
        """
        # First allow extensions to modify the data.
        for e in self._extensions:
            e.load(dump_data, record_cls)

        # Load explicitly defined model fields.
        model_data = {}
        it = self._model_fields.items()
        for model_field_name, (dump_key, dump_type) in it:
            model_data[model_field_name] = self._load_model_field(
                record_cls, model_field_name, dump_data, dump_key, dump_type)

        # Load model fields defined as system fields
        for systemfield in self._iter_modelfields(record_cls):
            model_data[systemfield.model_field_name] = self._load_model_field(
                record_cls, systemfield.model_field_name, dump_data,
                systemfield.dump_key, systemfield.dump_type)

        # Initialize model if an id was provided.
        if model_data.get('id') is not None:
            model_data['data'] = dump_data
            model = record_cls.model_cls(**model_data)
        else:
            model = None

        return record_cls(dump_data, model=model)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Base class for dumpers."""

from copy import deepcopy


class Dumper:
    """Interface for dumpers."""

    def dump(self, record, data):
        """Dump a record that can be used a source document for Elasticsearch.

        The job of this method is to create a Python dictionary from the record
        provided in the argument.

        If you overwrite this method without calling super, then you should
        ensure that you make a deep copy of the record dictionary, to avoid
        that changes to the dump affects the record.

        :param record: The record to dump.
        :param data: The initial dump data passed in by ``record.dumps()``.
        """
        data.update(deepcopy(dict(record)))
        return data

    def load(self, data, record_cls):
        """Load a record from the source document of an Elasticsearch hit.

        The job of this method, is to create a record of type ``record_cls``
        based on the input ``data``.

        :param data: A Python dictionary representing the data to load.
        :param records_cls: The record class to be constructed.
        :returns: A instance of ``record_cls``.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""System fields provides managed access to the record's dictionary.

**A simple example**

Take the following record subclass:

.. code-block:: python

    class MyRecord(Record, SystemFieldsMixin):
        test = ConstantField('mykey', 'myval')

The class defines a system field named ``test`` of the type ``ConstantField``.
The constant field adds a key (``mykey``) to the record with the value
``myval`` when a record is created:

.. code-block:: python

    record = MyRecord({})

The key ``mykey`` is part of the record's dictionary (i.e. you can do
``record['mykey']`` to acecss the value)::

    record['mykey'] == 'myval'

The key can however also be accessed through the field (i.e. ``record.test``)::

    record.test == 'myval'

System fields is thus a way to manage a subpart of record an allows you the
field to hook into the record API. This is a powerful API that can be used
to create fields which provides integration with related objects.

**A more advanced example**

Imagine the following record subclass using an imaginary ``PIDField``:

.. code-block:: python

    class MyRecord(Record, SystemFieldsMixin):
        pid = PIDField(pid_type='recid', object_type='rec')

You could use this field to set a PID on the record::

    record.pid = PersistentIdentifier(...)

Or, you could access the PID on a record you get from the database:

.. code-block:: python

    record = MyRecord.get_record()
    record.pid  # would return a PersistentIdentifier object.

The simple example only worked with the record itself. The more advanced
example here, the record is integrated with related objects.

**Data access layer**

System fields can do a lot, however you should seen them as part of the data
access layer. This means that they primarily simplifies data access between
records and related objects.
"""

from .base import SystemField, SystemFieldContext, SystemFieldsMeta, \
    SystemFieldsMixin
from .constant import ConstantField
from .dict import DictField
from .model import ModelField
from .relatedmodelfield import RelatedModelField, RelatedModelFieldContext
from .relations import ModelRelation, PKRelation, RelationsField

__all__ = (
    'ConstantField',
    'DictField',
    'ModelField',
    'ModelRelation',
    'PKRelation',
    'RelatedModelField',
    'RelatedModelFieldContext',
    'RelationsField',
    'SystemField',
    'SystemFieldContext',
    'SystemFieldsMeta',
    'SystemFieldsMixin',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field.

The field consists of:

- The system field itself.
- A relation mapping which is an object returned when accessing the system
  field from an instance (e.g. ``record.relations``).
- Relations (e.g. PKRelation) used for defining a relationship to a specific
  record type.
- Results, values returned when accessing a field (e.g.
  ``record.relations.languages``).
"""

from .errors import InvalidCheckValue, InvalidRelationValue, RelationError
from .field import RelationsField
from .mapping import RelationsMapping
from .modelrelations import ModelRelation
from .relations import ListRelation, NestedListRelation, PKListRelation, \
    PKNestedListRelation, PKRelation, RelationBase
from .results import RelationListResult, RelationNestedListResult, \
    RelationResult

__all__ = (
    'InvalidCheckValue',
    'InvalidRelationValue',
    'ListRelation',
    'ModelRelation',
    'NestedListRelation',
    'PKListRelation',
    'PKNestedListRelation',
    'PKRelation',
    'RelationBase',
    'RelationError',
    'RelationListResult',
    'RelationNestedListResult',
    'RelationResult',
    'RelationsField',
    'RelationsMapping',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records API."""


import copy
import uuid
from datetime import datetime, timedelta

import pytest
from jsonresolver import JSONResolver
from jsonresolver.contrib.jsonref import json_loader_factory
from jsonschema import FormatChecker
from jsonschema.exceptions import ValidationError
from sqlalchemy.orm.exc import NoResultFound

from invenio_records import Record
from invenio_records.errors import MissingModelError
from invenio_records.models import RecordMetadata
from invenio_records.validators import PartialDraft4Validator


def strip_ms(dt):
    """Strip microseconds."""
    return dt - timedelta(microseconds=dt.microsecond)


def test_get_records(testapp, db):
    """Test bulk record fetching."""
    # Create test records
    test_records = [
        Record.create({'title': 'test1'}),
        Record.create({'title': 'to_be_deleted'}),
        Record.create({'title': 'test3'}),
    ]
    db.session.commit()
    test_ids = [record.id for record in test_records]

    # Fetch test records
    assert len(Record.get_records(test_ids)) == 3

    test_records[1].delete()

    # should not show deleted
    db.session.commit()
    assert len(Record.get_records(test_ids)) == 2

    # should show deleted
    assert len(Record.get_records(test_ids, with_deleted=True)) == 3


def test_revision_id_created_updated_properties(testapp, db):
    """Test properties."""
    record = Record.create({'title': 'test'})
    assert record.revision_id == 0
    dt_c = record.created
    assert dt_c
    dt_u = record.updated
    assert dt_u
    record['title'] = 'test 2'
    record.commit()
    db.session.commit()
    assert record.revision_id == 1
    assert strip_ms(record.created) == strip_ms(dt_c)
    assert strip_ms(record.updated) >= strip_ms(dt_u)

    assert dt_u.tzinfo is None
    utcnow = datetime.utcnow()
    assert dt_u > utcnow - timedelta(seconds=10)
    assert dt_u < utcnow + timedelta(seconds=10)


def test_delete(testapp, database):
    """Test delete a record."""
    db = database
    # Create a record, revise it and delete it.
    record = Record.create({'title': 'test 1'})
    db.session.commit()
    record['title'] = 'test 2'
    record.commit()
    db.session.commit()
    record.delete()
    db.session.commit()

    # Deleted records a not retrievable by default
    pytest.raises(NoResultFound, Record.get_record, record.id)

    # Deleted records can be retrieved if you explicit request it
    record = Record.get_record(record.id, with_deleted=True)

    # Deleted records are empty
    assert record == {}
    assert record.model.json is None

    # Deleted records *cannot* be modified
    record['title'] = 'deleted'
    assert pytest.raises(MissingModelError, record.commit)

    # Deleted records *can* be reverted
    record = record.revert(-2)
    assert record['title'] == 'test 2'
    db.session.commit()

    # The "undeleted" record can now be retrieve again
    record = Record.get_record(record.id)
    assert record['title'] == 'test 2'

    # Force deleted record cannot be retrieved again
    record.delete(force=True)
    db.session.commit()
    pytest.raises(
        NoResultFound, Record.get_record, record.id,
        with_deleted=True)


def test_revisions(testapp, database):
    """Test revisions."""
    db = database
    # Create a record and make modifications to it.
    record = Record.create({'title': 'test 1'})
    rec_uuid = record.id
    db.session.commit()
    record['title'] = 'test 2'
    record.commit()
    db.session.commit()
    record['title'] = 'test 3'
    record.commit()
    db.session.commit()

    # Get the record
    record = Record.get_record(rec_uuid)
    assert record['title'] == 'test 3'
    assert record.revision_id == 2

    # Retrieve specific revisions
    rev1 = record.revisions[0]
    assert rev1['title'] == 'test 1'
    assert rev1.revision_id == 0

    rev2 = record.revisions[1]
    assert rev2['title'] == 'test 2'
    assert rev2.revision_id == 1

    # Latest revision is identical to record.
    rev_latest = record.revisions[-1]
    assert dict(rev_latest) == dict(record)

    # Revert to a specific revision
    record = record.revert(rev1.revision_id)
    assert record['title'] == 'test 1'
    assert record.created == rev1.created
    assert record.updated != rev1.updated
    assert record.revision_id == 3
    db.session.commit()

    # Get the record again and check it
    record = Record.get_record(rec_uuid)
    assert record['title'] == 'test 1'
    assert record.revision_id == 3

    # Make a change and ensure revision id is changed as well.
    record['title'] = 'modification'
    record.commit()
    db.session.commit()
    assert record.revision_id == 4

    # Iterate over revisions
    assert len(record.revisions) == 5
    revs = list(record.revisions)
    assert revs[0]['title'] == 'test 1'
    assert revs[1]['title'] == 'test 2'
    assert revs[2]['title'] == 'test 3'
    assert revs[3]['title'] == 'test 1'
    assert revs[4]['title'] == 'modification'

    assert 2 in record.revisions
    assert 5 not in record.revisions


def test_retrieve_proper_revision(testapp, database):
    """Check accessing revision with gaps

    Test checks if it's possible to access proper revision
    when revision numbers have 'gaps'
    """
    db = database
    record = Record.create({'title': 'test 1'})
    db.session.commit()
    record['title'] = "test 2"
    record.commit()
    record['title'] = "test 3"
    record.commit()
    db.session.commit()
    record['title'] = "test 4"
    record.commit()
    db.session.commit()
    record['title'] = "test 5"
    record.commit()
    db.session.commit()

    assert len(record.revisions) == 4
    revs = list(record.revisions)
    assert revs[0]['title'] == "test 1"
    assert revs[0].revision_id == 0
    assert revs[1]['title'] == "test 3"
    assert revs[1].revision_id == 2
    assert revs[2]['title'] == "test 4"
    assert revs[2].revision_id == 3
    assert revs[3]['title'] == "test 5"
    assert revs[3].revision_id == 4

    # Access revision by revision_id
    rev_2 = record.revisions[2]
    assert rev_2['title'] == "test 3"
    assert rev_2.revision_id == 2

    # Access revision by negative list index
    assert rev_2 == record.revisions[-3]


def test_record_update_mutable(testapp, database):
    """Test updating mutables in a record."""
    db = database
    recid = uuid.UUID('262d2748-ba41-456f-a844-4d043a419a6f')

    # Create a new record with two mutables, a list and a dict
    rec = Record.create(
        {
            'title': 'Title',
            'list': ['foo', ],
            'dict': {'moo': 'boo'},
        },
        id_=recid)
    # Make sure mutables are there before and after commit
    assert rec == {
        'title': 'Title',
        'list': ['foo', ],
        'dict': {'moo': 'boo'}
    }
    db.session.commit()
    db.session.expunge_all()
    rec = Record.get_record(recid)
    assert rec == {
        'title': 'Title',
        'list': ['foo', ],
        'dict': {'moo': 'boo'}
    }

    # Set the mutables under key
    rec['list'] = ['bar', ]
    rec['dict'] = {'eggs': 'bacon'}
    rec.commit()
    # Make sure it commits to DB
    assert rec == {
        'title': 'Title',
        'list': ['bar', ],
        'dict': {'eggs': 'bacon'}
    }
    db.session.commit()
    db.session.expunge_all()
    rec = Record.get_record(recid)
    assert rec == {
        'title': 'Title',
        'list': ['bar', ],
        'dict': {'eggs': 'bacon'}
    }

    # Update the mutables under key
    rec['list'].append('spam')
    rec['dict']['ham'] = 'chicken'
    rec.commit()
    # Make sure it commits to DB
    assert rec == {
        'title': 'Title',
        'list': ['bar', 'spam'],
        'dict': {'eggs': 'bacon', 'ham': 'chicken'}
    }
    db.session.commit()
    db.session.expunge_all()
    rec = Record.get_record(recid)
    assert rec == {
        'title': 'Title',
        'list': ['bar', 'spam'],
        'dict': {'eggs': 'bacon', 'ham': 'chicken'}
    }


def test_missing_model(testapp, db):
    """Test revisions."""
    record = Record({})
    assert record.id is None
    assert record.revision_id is None
    assert record.created is None
    assert record.updated is None

    try:
        record.revisions
        assert False
    except MissingModelError:
        assert True

    pytest.raises(MissingModelError, record.commit)
    pytest.raises(MissingModelError, record.delete)
    pytest.raises(MissingModelError, record.revert, -1)


def test_record_replace_refs(testapp, db):
    """Test the replacement of JSON references using JSONResolver."""
    record = Record.create({
        'one': {'$ref': 'http://nest.ed/A'},
        'three': {'$ref': 'http://nest.ed/ABC'}
    })
    testapp.extensions['invenio-records'].loader_cls = json_loader_factory(
        JSONResolver(plugins=['demo.json_resolver']))

    record.enable_jsonref = False
    assert record.replace_refs() == record

    record.enable_jsonref = True
    out_json = record.replace_refs()
    assert out_json != record

    expected_json = {
        'one': {
            'letter': 'A',
            'next': '.',
        },
        'three': {
            'letter': 'A',
            'next': {
                'letter': 'B',
                'next': {
                    'letter': 'C',
                    'next': '.',
                },
            },
        },
    }
    assert out_json == expected_json


def test_replace_refs_deepcopy(testapp):
    """Test problem with replace_refs and deepcopy."""
    with testapp.app_context():
        assert copy.deepcopy(Record({'recid': 1}).replace_refs()) \
            == {'recid': 1}


def test_record_dump(testapp, db):
    """Test record dump method."""
    with testapp.app_context():
        record = Record.create({'foo': {'bar': 'Bazz', }, })
        record_dump = record.dumps()
        record_dump['foo']['bar'] = 'Spam'
        assert record_dump['foo']['bar'] != record['foo']['bar']


def test_default_format_checker(testapp):
    """Test a default format checker."""
    checker = FormatChecker()
    checker.checks('foo')(lambda el: el.startswith('foo'))
    data = {
        # The key 'bar' will fail validation if the format checker is used.
        'bar': 'bar',
        '$schema': {
            'properties': {
                'bar': {'format': 'foo'}
            }
        }
    }

    class CustomRecord(Record):
        format_checker = checker

    assert pytest.raises(ValidationError, CustomRecord(data).validate)


def test_validate_with_format(testapp, db):
    """Test that validation can accept custom format rules."""
    with testapp.app_context():
        checker = FormatChecker()
        checker.checks('foo')(lambda el: el.startswith('foo'))
        data = {
            'bar': 'foo',
            '$schema': {
                'properties': {
                    'bar': {'format': 'foo'}
                }
            }
        }

        # test record creation with valid data
        assert data == Record.create(data)
        record = Record.create(data, format_checker=checker)
        # test direct call to validate with valid data
        assert record.validate(format_checker=checker) is None
        # test commit with valid data
        record.commit(format_checker=checker)

        record['bar'] = 'bar'
        # test direct call to validate with invalid data
        with pytest.raises(ValidationError) as excinfo:
            record.validate(format_checker=checker)
        assert "'bar' is not a 'foo'" in str(excinfo.value)
        # test commit with invalid data
        with pytest.raises(ValidationError) as excinfo:
            record.commit(format_checker=checker)
        assert "'bar' is not a 'foo'" in str(excinfo.value)

        data['bar'] = 'bar'
        # test record creation with invalid data
        with pytest.raises(ValidationError) as excinfo:
            record = Record.create(data, format_checker=checker)
        assert "'bar' is not a 'foo'" in str(excinfo.value)


def test_validate_partial(testapp, db):
    """Test partial validation."""
    schema = {
        'properties': {
            'a': {'type': 'string'},
            'b': {'type': 'string'},
        },
        'required': ['b']
    }
    data = {
        'a': 'hello',
        '$schema': schema
    }
    with testapp.app_context():
        # Test validation on create()

        # normal validation should fail because 'b' is required
        with pytest.raises(ValidationError) as exc_info:
            Record.create(data)
        assert "'b' is a required property" == exc_info.value.message
        # validate with a less restrictive validator
        record = Record.create(data, validator=PartialDraft4Validator)
        # set wrong data types should fails in any case
        data_incorrect = copy.deepcopy(data)
        data_incorrect['a'] = 1
        with pytest.raises(ValidationError) as exc_info:
            Record.create(data_incorrect, validator=PartialDraft4Validator)
        assert "1 is not of type 'string'" == exc_info.value.message

        # Test validation on commit()

        # validation not passing with normal validator
        with pytest.raises(ValidationError) as exc_info:
            record.commit()
        assert "'b' is a required property" == exc_info.value.message
        # validation passing with less restrictive validator
        assert data == record.commit(validator=PartialDraft4Validator)
        # set wrong data types should fails in any case
        record['a'] = 1
        with pytest.raises(ValidationError) as exc_info:
            record.commit(validator=PartialDraft4Validator)


def test_reversed_works_for_revisions(testapp, database):
    db = database
    record = Record.create({'title': 'test 1'})
    db.session.commit()

    record['title'] = "test 2"
    record.commit()
    record['title'] = "Test 3"
    db.session.commit()

    record['title'] = "test 4"
    record.commit()
    db.session.commit()

    reversed_revisions = list(reversed(record.revisions))
    reversed_revisions[0].revision_id == 3
    reversed_revisions[1].revision_id == 1
    reversed_revisions[2].revision_id == 0


def test_clear_none(testapp, db):
    """Test clear_none."""
    record = Record({'a': None})
    record.clear_none()
    assert record == {}


def test_undelete_no_get(testapp, db):
    """Test undelete a record."""
    record = Record.create({'title': 'test'})
    db.session.commit()
    record.delete()
    db.session.commit()
    record.undelete()
    record.commit()
    db.session.commit()
    assert record == {'title': 'test'}


def test_undelete_with_get(testapp, db):
    """Test undelete a record."""
    record = Record.create({'title': 'test'})
    db.session.commit()
    record.delete()
    db.session.commit()
    record = Record.get_record(record.id, with_deleted=True)
    record.undelete()
    record.commit()
    db.session.commit()
    assert record == {}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2021 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for metadata storage."""


from functools import lru_cache

from invenio_base.utils import obj_or_import_string
from jsonref import JsonRef
from jsonresolver import JSONResolver
from jsonresolver.contrib.jsonref import json_loader_factory
from jsonresolver.contrib.jsonschema import ref_resolver_factory
from jsonschema import validate

from invenio_records.errors import RecordsRefResolverConfigError
from invenio_records.resolver import urljoin_with_custom_scheme

from . import config
from .validators import _create_validator


class _RecordsState(object):
    """State for record JSON resolver."""

    def __init__(self, app, entry_point_group=None):
        """Initialize state."""
        self.app = app
        self.resolver = JSONResolver(entry_point_group=entry_point_group)
        self.refresolver_cls = ref_resolver_factory(self.resolver)
        self.refresolver_store = None
        if (
            self.app.config.get("RECORDS_REFRESOLVER_CLS")
        ):
            self.refresolver_cls = obj_or_import_string(
                self.app.config.get("RECORDS_REFRESOLVER_CLS"),
            )
            self.refresolver_store = obj_or_import_string(
                self.app.config.get("RECORDS_REFRESOLVER_STORE")
            )

        self.loader_cls = json_loader_factory(self.resolver)

    def validate(self, data, schema, **kwargs):
        """Validate data using schema with ``JSONResolver``."""
        if not isinstance(schema, dict):
            schema = {'$ref': schema}
        refresolver_cls_kwargs = {}
        if self.refresolver_store:
            refresolver_cls_kwargs['store'] = self.refresolver_store
            refresolver_cls_kwargs['urljoin_cache'] = \
                lru_cache(1024)(urljoin_with_custom_scheme)

        validator_cls = _create_validator(
            schema=schema,
            base_validator_cls=kwargs.pop("cls", None),
            custom_checks=self.app.config.get('RECORDS_VALIDATION_TYPES', {})
        )

        resolver = self.refresolver_cls.from_schema(
            schema,
            **refresolver_cls_kwargs
        )

        return validate(
            data,
            schema,
            cls=validator_cls,
            resolver=resolver,
            **kwargs
        )

    def replace_refs(self, data):
        """Replace the JSON reference objects with ``JsonRef``."""
        return JsonRef.replace_refs(data, loader=self.loader_cls())


class InvenioRecords(object):
    """Invenio-Records extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization."""
        if app:
            self._state = self.init_app(app, **kwargs)

    def init_app(self, app,
                 entry_point_group='invenio_records.jsonresolver', **kwargs):
        """Flask application initialization.

        :param app: The Flask application.
        :param entry_point_group: The entrypoint for jsonresolver extensions.
            (Default: ``'invenio_records.jsonresolver'``)
        """
        self.init_config(app)
        state = _RecordsState(app, entry_point_group=entry_point_group)
        app.extensions['invenio-records'] = state
        return state

    def init_config(self, app):
        """Initialize configuration.

        :param app: The Flask application.
        """
        app.config.setdefault('SQLALCHEMY_TRACK_MODIFICATIONS', True)
        for k in dir(config):
            if k.startswith('RECORDS_'):
                app.config.setdefault(k, getattr(config, k))

        if (
            app.config.get("RECORDS_REFRESOLVER_CLS")
            and not app.config.get("RECORDS_REFRESOLVER_STORE")
        ):
            raise RecordsRefResolverConfigError(
                "RECORDS_REFRESOLVER_CLS config requires "
                "RECORDS_REFRESOLVER_STORE to be set."
            )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test of dictionary utilities."""

from copy import deepcopy

import pytest

from invenio_records.dictutils import clear_none, dict_lookup


def test_clear_none():
    """Test clearning of the dictionary."""
    d = {
        'a': None,
        'b': {
            'c': None
        },
        'd': ['1', None, []],
        'e': [{'a': None, 'b': []}],
    }

    clear_none(d)
    # Modifications are done in place, so gotta test after the function call.
    assert d == {'d': ['1']}

    d = {
        'a': None,
        'b': [
            {'a': '1', 'b': None},
        ],
    }
    clear_none(d)
    # Modifications are done in place, so gotta test after the function call.
    assert d == {'b': [{'a': '1'}]}


def test_dict_lookup():
    """Test lookup by a key."""
    d = {
        'a': 1,
        'b': {
            'c': None
        },
        'd': ['1', '2'],
    }
    assert dict_lookup(d, 'a') == d['a']
    assert dict_lookup(d, 'b') == d['b']
    assert dict_lookup(d, 'b.c') == d['b']['c']
    assert dict_lookup(d, 'd') == d['d']
    assert dict_lookup(d, 'd.0') == d['d'][0]
    assert dict_lookup(d, 'd.1') == d['d'][1]
    assert dict_lookup(d, 'd.-1') == d['d'][-1]

    assert pytest.raises(KeyError, dict_lookup, d, 'x')
    assert pytest.raises(KeyError, dict_lookup, d, 'a.x')
    assert pytest.raises(KeyError, dict_lookup, d, 'b.x')
    assert pytest.raises(KeyError, dict_lookup, d, 'b.c.0')
    assert pytest.raises(KeyError, dict_lookup, d, 'd.3')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record models."""

import uuid
from copy import deepcopy
from datetime import datetime

from invenio_db import db
from sqlalchemy.dialects import mysql, postgresql
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy_utils.types import JSONType, UUIDType


class Timestamp(object):
    """Timestamp model mix-in with fractional seconds support.

    SQLAlchemy-Utils timestamp model does not have support for fractional
    seconds.
    """

    created = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
        default=datetime.utcnow,
        nullable=False
    )
    updated = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
        default=datetime.utcnow,
        nullable=False
    )


@db.event.listens_for(Timestamp, 'before_update', propagate=True)
def timestamp_before_update(mapper, connection, target):
    """Update `updated` property with current time on `before_update` event."""
    target.updated = datetime.utcnow()


class RecordMetadataBase(Timestamp):
    """Represent a base class for record metadata.

    The RecordMetadata object contains a ``created`` and  a ``updated``
    properties that are automatically updated.
    """

    encoder = None
    """"Class-level attribute to set a JSON data encoder/decoder.

    This allows customizing you to e.g. convert specific entries to complex
    Python objects. For instance you could convert ISO-formatted datetime
    objects into Python datetime objects.
    """

    id = db.Column(
        UUIDType,
        primary_key=True,
        default=uuid.uuid4,
    )
    """Record identifier."""

    json = db.Column(
        db.JSON().with_variant(
            postgresql.JSONB(none_as_null=True),
            'postgresql',
        ).with_variant(
            JSONType(),
            'sqlite',
        ).with_variant(
            JSONType(),
            'mysql',
        ),
        default=lambda: dict(),
        nullable=True
    )
    """Store metadata in JSON format.

    When you create a new ``Record`` the ``json`` field value should never be
    ``NULL``. Default value is an empty dict. ``NULL`` value means that the
    record metadata has been deleted.
    """

    # Enables SQLAlchemy version counter (not the same as SQLAlchemy-Continuum)
    version_id = db.Column(db.Integer, nullable=False)
    """Used by SQLAlchemy for optimistic concurrency control."""

    __mapper_args__ = {
        'version_id_col': version_id
    }

    def __init__(self, data=None, **kwargs):
        """Initialize the model specifically by setting the."""
        if data is not None:
            self.data = data
        super(RecordMetadataBase, self).__init__(self, **kwargs)

    @hybrid_property
    def is_deleted(self):
        """Boolean flag to determine if a record is soft deleted."""
        return self.json == None  # noqa

    @is_deleted.setter
    def is_deleted(self, value):
        """Boolean flag to set record as soft deleted.

        This propert sets the JSON colum to None. The hybrid property *cannot*
        be used to undelete a record by setting the property to False.
        """
        if value is True:
            self.json = None
        else:
            self.json = {}

    @property
    def data(self):
        """Get data by decoding the JSON.

        This allows a subclass to override
        """
        # We make a deepcopy in order to completely disconnect updates on the
        # record dict from the model's JSON. Otherwise changes made by the
        # encoder/decode, and updates by users are propagated to the model's
        # json field (circumventing the encoder) and likely causing the JSON
        # serialization errors when saving to the DB.
        return self.decode(self.json)

    @data.setter
    def data(self, value):
        """Set data by encoding the JSON.

        This allows a subclass to override
        """
        self.json = self.encode(value)
        flag_modified(self, 'json')

    @classmethod
    def encode(cls, value):
        """Encode a JSON document."""
        data = deepcopy(value)
        return cls.encoder.encode(data) if cls.encoder else data

    @classmethod
    def decode(cls, json):
        """Decode a JSON document."""
        data = deepcopy(json)
        return cls.encoder.decode(data) if cls.encoder else data


class RecordMetadata(db.Model, RecordMetadataBase):
    """Represent a record metadata."""

    __tablename__ = 'records_metadata'

    # Enables SQLAlchemy-Continuum versioning
    __versioned__ = {}


__all__ = (
    'RecordMetadata',
    'RecordMetadataBase',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records."""


import json
import os
import uuid

import pytest
from flask import Flask
from invenio_db.utils import drop_alembic_version_table
from jsonschema.exceptions import ValidationError
from sqlalchemy.orm.exc import NoResultFound

from invenio_records import InvenioRecords, Record
from invenio_records.errors import MissingModelError


def test_version():
    """Test version import."""
    from invenio_records import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = InvenioRecords(app)
    assert 'invenio-records' in app.extensions

    app = Flask('testapp')
    ext = InvenioRecords()
    assert 'invenio-records' not in app.extensions
    ext.init_app(app)
    assert 'invenio-records' in app.extensions


def test_db(testapp, db):
    """Test database backend."""
    with testapp.app_context():
        assert 'records_metadata' in db.metadata.tables
        assert 'records_metadata_version' in db.metadata.tables
        assert 'transaction' in db.metadata.tables

    schema = {
        'type': 'object',
        'properties': {
            'title': {'type': 'string'},
            'field': {'type': 'boolean'},
            'hello': {'type': 'array'},
        },
        'required': ['title'],
    }
    data = {'title': 'Test', '$schema': schema}
    from invenio_records.models import RecordMetadata as RM

    # Create a record
    with testapp.app_context():
        assert RM.query.count() == 0

        record_uuid = Record.create(data).id
        db.session.commit()

        assert RM.query.count() == 1
        db.session.commit()

    # Retrieve created record
    with testapp.app_context():
        record = Record.get_record(record_uuid)
        assert record.dumps() == data
        with pytest.raises(NoResultFound):
            Record.get_record(uuid.uuid4())
        record['field'] = True
        record = record.patch([
            {'op': 'add', 'path': '/hello', 'value': ['world']}
        ])
        assert record['hello'] == ['world']
        record.commit()
        db.session.commit()

    with testapp.app_context():
        record2 = Record.get_record(record_uuid)
        assert record2.model.version_id == 2
        assert record2['field']
        assert record2['hello'] == ['world']
        db.session.commit()

    # Cannot commit record without model (i.e. Record.create_record)
    with testapp.app_context():
        record3 = Record({'title': 'Not possible'})
        with pytest.raises(MissingModelError):
            record3.commit()

    # Check invalid schema values
    with testapp.app_context():
        data = {
            '$schema': 'http://json-schema.org/learn/examples/'
                       'geographical-location.schema.json',
            'latitude': 42,
            'longitude': 42,
        }

        record_with_schema = Record.create(data).commit()
        db.session.commit()

        record_with_schema['latitude'] = 'invalid'
        with pytest.raises(ValidationError):
            record_with_schema.commit()

    # Allow types overriding on schema validation
    with testapp.app_context():
        data = {
            'title': 'Test',
            'hello': tuple(['foo', 'bar']),
            '$schema': schema
        }
        testapp.config['RECORDS_VALIDATION_TYPES'] = {}
        with pytest.raises(ValidationError):
            Record.create(data).commit()

        testapp.config['RECORDS_VALIDATION_TYPES'] = {'array': (list, tuple)}
        record_uuid = Record.create(data).commit()
        db.session.commit()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create records tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = '862037093962'
down_revision = '1095cdf9f350'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        'records_metadata',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column(
            'id', sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column('json', sqlalchemy_utils.JSONType().with_variant(
            sa.dialects.postgresql.JSON(
                none_as_null=True), 'postgresql',
        ), nullable=True),
        sa.Column('version_id', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'records_metadata_version',
        sa.Column('created', sa.DateTime(),
                  autoincrement=False, nullable=True),
        sa.Column('updated', sa.DateTime(),
                  autoincrement=False, nullable=True),
        sa.Column('id', sqlalchemy_utils.types.uuid.UUIDType(),
                  autoincrement=False, nullable=False),
        sa.Column('json', sqlalchemy_utils.JSONType().with_variant(
            sa.dialects.postgresql.JSON(
                none_as_null=True), 'postgresql',
        ), autoincrement=False, nullable=True),
        sa.Column('version_id', sa.Integer(),
                  autoincrement=False, nullable=True),
        sa.Column('transaction_id', sa.BigInteger(),
                  autoincrement=False, nullable=False),
        sa.Column('end_transaction_id',
                  sa.BigInteger(), nullable=True),
        sa.Column('operation_type',
                  sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint('id', 'transaction_id')
    )
    op.create_index(
        op.f('ix_records_metadata_version_end_transaction_id'),
        'records_metadata_version', ['end_transaction_id'], unique=False
    )
    op.create_index(
        op.f('ix_records_metadata_version_operation_type'),
        'records_metadata_version', ['operation_type'], unique=False
    )
    op.create_index(
        op.f('ix_records_metadata_version_transaction_id'),
        'records_metadata_version', ['transaction_id'], unique=False
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(op.f('ix_records_metadata_version_transaction_id'),
                  table_name='records_metadata_version')
    op.drop_index(op.f('ix_records_metadata_version_operation_type'),
                  table_name='records_metadata_version')
    op.drop_index(op.f('ix_records_metadata_version_end_transaction_id'),
                  table_name='records_metadata_version')
    op.drop_table('records_metadata_version')
    op.drop_table('records_metadata')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Alter column from json to jsonb."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '07fb52561c5c'
down_revision = '862037093962'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    if op._proxy.migration_context.dialect.name == 'postgresql':
        op.alter_column(
            'records_metadata',
            'json',
            type_=sa.dialects.postgresql.JSONB,
            postgresql_using='json::text::jsonb'
        )


def downgrade():
    """Downgrade database."""
    if op._proxy.migration_context.dialect.name == 'postgresql':
        op.alter_column(
            'records_metadata',
            'json',
            type_=sa.dialects.postgresql.JSON,
            postgresql_using='json::text::json'
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Model relations.

A model relation relies on database-level foreign keys to define a relationship
instead of injecting the IDs inside the record dictionary. On indexing, the
specified attributes from the related record are dumped into the index.

.. code-block:: python

    # Your database model defines a foreign key to the related model.
    class MyRecordModel(db.Model, RecordMetadataBase):
        # FK:
        user_id = db.Column(UserModel.id, ..., db.ForeignKey(...))

    class MyRecord(Record)

        # The database model used by the record.
        model_cls = MyRecordModel

        # System field defining the related_id as a model attribute (and
        # ensuring the id is dumped to the index).
        user_id = ModelField("user_id")

        # System field defining the relation
        relations = RelationsField(
            related=ModelRelation(
                # The related record class.
                UserRecord,
                # The attribute name of the ModelField.
                'user_id',
                # Top-level key to dump in index
                'user',
                # Attributes to dump (in addition to id)
                keys=['email', 'username', 'profile'],
            )
"""

from .errors import InvalidRelationValue
from .relations import RelationBase
from .results import RelationResult


class ModelRelationResult(RelationResult):
    """Result class for a model relation."""

    def _lookup_id(self):
        return getattr(self.record, self.field._model_field_name)

    def validate(self):
        """Validate relation."""
        # The relation is validated via database-level foreign key constraints.
        return True

    def dereference(self, keys=None, attrs=None):
        """Dereference the relation field object inside the record."""
        # Prepare the record dictionary to look like: {'id': ...}
        if self.field.key not in self.record:
            id_ = self._lookup_id()
            if id_ is None:
                return None
            else:
                self.record[self.field.key] = {
                    self.field._value_key_suffix: str(id_)
                }

        data = self.record[self.field.key]
        return self._dereference_one(
            data, keys or self.keys, attrs or self.attrs)

    def clean(self, keys=None, attrs=None):
        """Clean the record."""
        # Don't store anything inside the record committed to the database.
        self.record.pop(self.field.key, None)


class ModelRelation(RelationBase):
    """Define a relation stored as a foreign key on the record's model."""

    result_cls = ModelRelationResult

    def __init__(self, record_cls, model_field_name, key, keys=None,
                 attrs=None):
        """Constructor."""
        self._record_cls = record_cls
        self._model_field_name = model_field_name
        super().__init__(key=key, keys=keys, attrs=attrs)

    def resolve(self, id_):
        """Get the related object."""
        return self._record_cls.get_record(id_)

    def parse_value(self, value):
        """Extract id from object being set on relation."""
        if isinstance(value, self._record_cls):
            return value.id
        elif isinstance(value, self._record_cls.model_cls):
            return value.id
        else:
            raise InvalidRelationValue(
                f'Invalid value. Expected {self._record_cls} or '
                f'{self._record_cls.model_cls}'
            )

    def set_value(self, record, value):
        """Set an related object."""
        store_value = self.parse_value(value)
        # The existence of the related object is checked via database-level
        # foreign key constraints.
        setattr(record, self._model_field_name, store_value)

    def get_value(self, record):
        """Return the resolved relation from a record."""
        return self.result_cls(self, record)

    def clear_value(self, record):
        """Clear the relation."""
        setattr(record, self._model_field_name, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for relations system field."""

from collections.abc import Iterable

import pytest

from invenio_records.api import Record
from invenio_records.systemfields import PKRelation, RelationsField, \
    SystemFieldsMixin
from invenio_records.systemfields.relations import InvalidRelationValue, \
    PKListRelation, PKNestedListRelation, RelationsMapping
from invenio_records.systemfields.relations.errors import InvalidCheckValue


@pytest.fixture()
def languages(db):
    class Language(Record, SystemFieldsMixin):
        pass

    languages_data = (
        {'title': 'English', 'iso': 'en', 'information': {
            'native_speakers': '400 million', 'ethnicity': 'English'}},
        {'title': 'French', 'iso': 'fr', 'information': {
            'native_speakers': '76.8 million', 'ethnicity': 'French'}},
        {'title': 'Spanish', 'iso': 'es', 'information': {
            'native_speakers': '489 million', 'ethnicity': 'Spanish'}},
        {'title': 'Italian', 'iso': 'it', 'information': {
            'native_speakers': '67 million', 'ethnicity': 'Italians'}},
        {'title': 'Old English', 'iso': 'oe', 'information': {
            'native_speakers': '400 million',
            'ethnicity': ['English', 'Old english']}},
    )

    languages = {}
    for lang in languages_data:
        lang_rec = Language.create(lang)
        languages[lang['iso']] = lang_rec
    db.session.commit()
    return Language, languages


def test_relations_field_pk_relation(testapp, db, languages):
    """RelationsField tests for PKRelation."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key='language', keys=['iso'], record_cls=Language),
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.language, PKRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.language)
    assert record.relations.language() is None

    # Initialize from dictionary data
    record = Record1({'language': {'id': str(en_lang.id)}})
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    # Initialize from field data
    record = Record1({}, relations={'language': str(en_lang.id)})
    assert record['language']['id'] == str(en_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    # Set via record data
    record = Record1.create({})
    record['language'] = {'id': str(en_lang.id)}
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang
    record.commit()  # validates
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    record['language'] = {'id': 'invalid'}
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.language = str(en_lang.id)
    assert record['language']['id'] == str(en_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    # Update existing value
    record.relations.language = str(fr_lang.id)
    assert record['language']['id'] == str(fr_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.language = 'invalid'
    # Check that old value is still there
    assert record['language']['id'] == str(fr_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    # Dereference relation
    record.relations.language.dereference()
    assert record['language'] == {
        'id': str(fr_lang.id),
        'iso': 'fr',  # only stores the "iso" field
        '@v': str(fr_lang.id) + '::' + str(fr_lang.revision_id),
    }

    # Commit clean dereferened fields
    record.commit()
    assert record['language'] == {'id': str(fr_lang.id)}

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record['language'] == {'id': str(fr_lang.id)}
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    # Clear field
    record.relations.language = None
    assert 'language' not in record
    assert record.relations.language() is None
    record.commit()


def test_relations_field_pk_list_relation(testapp, db, languages):
    """RelationsField tests for PKListRelation."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key='languages', keys=['iso', 'information.ethnicity'],
                record_cls=Language),
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.languages, PKListRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.languages)
    assert record.relations.languages() is None

    # Initialize from dictionary data
    record = Record1({'languages': [{'id': str(en_lang.id)}]})
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Initialize from field data
    record = Record1({}, relations={'languages': [str(en_lang.id)]})
    assert record['languages'] == [{'id': str(en_lang.id)}]
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Set via record dictionary data
    record = Record1.create({})
    record['languages'] = [{'id': str(en_lang.id)}]
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    record.commit()  # validates
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    record['languages'] = {'id': 'invalid'}
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.languages = [str(en_lang.id)]
    assert record['languages'][0]['id'] == str(en_lang.id)
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Update existing value
    record.relations.languages = [str(fr_lang.id)]
    assert record['languages'][0]['id'] == str(fr_lang.id)
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    # Set invalid value
    for v in ('invalid', ['invalid']):
        with pytest.raises(InvalidRelationValue):
            record.relations.languages = v
        # Check that old value is still there
        assert record['languages'][0]['id'] == str(fr_lang.id)
        res_iter = record.relations.languages()
        assert isinstance(res_iter, Iterable)
        res = next(res_iter)
        assert res == fr_lang

    # Dereference relation
    record.relations.languages.dereference()
    assert record['languages'] == [{
        'id': str(fr_lang.id),
        'iso': 'fr',  # only stores the "iso" field
        'information': {'ethnicity': 'French'},
        '@v': str(fr_lang.id) + '::' + str(fr_lang.revision_id),
    }]

    # Commit clean dereferened fields
    record.commit()
    assert record['languages'] == [{'id': str(fr_lang.id)}]

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record['languages'] == [{'id': str(fr_lang.id)}]
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    # Clear field
    record.relations.languages = None
    assert 'languages' not in record
    assert record.relations.languages() is None
    record.commit()


def test_relations_field_pk_list_relation_of_objects(testapp, db, languages):
    """RelationsField tests for PKListRelation with a list of objects."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            array_of_objects=PKListRelation(
                key='array_of_objects',
                keys=['iso'],
                record_cls=Language,
                relation_field='language'
            ),
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.array_of_objects, PKListRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.array_of_objects)
    assert record.relations.array_of_objects() is None

    # Initialize from dictionary data
    record = Record1({'array_of_objects': [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            }
        },
        {
            'language': {"id": str(en_lang.id)},
            'field1': "string"
        },
        {
            'language': {"id": str(fr_lang.id)},
            'field2': "string2"
        },
        {
            'language': {"id": str(en_lang.id)},
            'field1': "string"
        },
    ]})

    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang
    res = next(res_iter)
    assert res == en_lang

    # Initialize from field data
    record = Record1(
        {"metadata": {"title": "random title"}},
        relations={'array_of_objects': [
            {
                'field1': "no language",
                'field2': "random string",
                'field3': {
                    'nestedField': 'nested string'
                }
            },
            {
                'language': str(en_lang.id),
                'field1': "string"
            },
            {
                'language': str(fr_lang.id),
                'field2': "string2"
            },
            {
                'language': str(en_lang.id),
                'field1': "string"
            },
        ]})
    with pytest.raises(KeyError):
        record['array_of_objects'][0]['language']['id']
    assert record['array_of_objects'][1]['language']['id'] == str(
        en_lang.id)
    assert record['array_of_objects'][2]['language']['id'] == str(
        fr_lang.id)
    assert record['array_of_objects'][3]['language']['id'] == str(
        en_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang
    res = next(res_iter)
    assert res == en_lang

    # Set via record dictionary data
    record = Record1.create({})
    record['array_of_objects'] = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            }
        },
        {
            'language': {"id": str(en_lang.id)},
            'field1': "string"
        },
        {
            'language': {"id": str(fr_lang.id)},
            'field2': "string2"
        },
        {
            'language': {"id": str(en_lang.id)},
            'field1': "string"
        },
    ]
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang

    record.commit()  # validates
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang

    record['array_of_objects'] = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            },
            'language': {'id': 'invalid'}
        },
        {'field1': "testString"}
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.array_of_objects = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            }
        },
        {
            'language': str(en_lang.id),
            'field1': "string"
        },
        {
            'language': str(fr_lang.id),
            'field2': "string2"
        },
        {
            'language': str(en_lang.id),
            'field1': "string"
        },
    ]

    assert record['array_of_objects'][1]['language']['id'] == str(en_lang.id)
    assert record['array_of_objects'][2]['language']['id'] == str(fr_lang.id)
    assert record['array_of_objects'][3]['language']['id'] == str(en_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang
    res = next(res_iter)
    assert res == en_lang

    # Update existing value
    record.relations.array_of_objects = [
        {
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            },
            'language': str(en_lang.id),

        },
        {
            'language': str(es_lang.id),
            'field1': "string"
        },
        {
            'language': str(en_lang.id),
            'field1': "string"
        },
    ]
    assert record['array_of_objects'][0]['language']['id'] == str(en_lang.id)
    assert record['array_of_objects'][1]['language']['id'] == str(es_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == es_lang

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.array_of_objects = 'invalid'
    # Check that old value is still there
    assert record['array_of_objects'][0]['language']['id'] == str(en_lang.id)
    assert record['array_of_objects'][1]['language']['id'] == str(es_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == es_lang

    # Dereference relation
    record.relations.array_of_objects.dereference()
    assert record['array_of_objects'] == [
        {
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            },
            'language': {
                'id': str(en_lang.id),
                'iso': 'en',
                '@v': str(en_lang.id) + '::' + str(en_lang.revision_id),
            }

        },
        {
            'language': {
                'id': str(es_lang.id),
                'iso': 'es',
                '@v': str(es_lang.id) + '::' + str(es_lang.revision_id),
            },
            'field1': "string"
        },
        {
            'language': {
                'id': str(en_lang.id),
                'iso': 'en',
                '@v': str(en_lang.id) + '::' + str(en_lang.revision_id),
            },
            'field1': "string"
        },
    ]

    # Commit clean dereferenced fields
    record.commit()
    assert record['array_of_objects'][0]['language']['id'] == str(en_lang.id)
    assert record['array_of_objects'][1]['language']['id'] == str(es_lang.id)
    assert record['array_of_objects'][2]['language']['id'] == str(en_lang.id)

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record['array_of_objects'][0]['language']['id'] == str(en_lang.id)
    assert record['array_of_objects'][1]['language']['id'] == str(es_lang.id)
    assert record['array_of_objects'][2]['language']['id'] == str(en_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == es_lang
    res = next(res_iter)
    assert res == en_lang

    # Clear field
    record.relations.array_of_objects = None
    assert 'array_of_objects' not in record
    assert record.relations.array_of_objects() is None
    record.commit()


def test_nested_relations_field(testapp, db, languages):
    """Tests deeply nested relations."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']
    it_lang = languages['it']

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            deep_single=PKRelation(
                key='metadata.deep.language', record_cls=Language),
            deep_list=PKListRelation(
                key='metadata.deep.languages', record_cls=Language),
            deep_sibling=PKRelation(
                key='metadata.deep.dark.language', record_cls=Language),
        )

    # Set all fields
    record = Record1.create({})
    record.relations.deep_single = en_lang
    record.relations.deep_list = [en_lang, fr_lang, es_lang]
    record.relations.deep_sibling = it_lang

    assert record == {
        'metadata': {
            'deep': {
                'language': {'id': str(en_lang.id)},
                'languages': [
                    {'id': str(en_lang.id)},
                    {'id': str(fr_lang.id)},
                    {'id': str(es_lang.id)}
                ],
                'dark': {
                    'language': {'id': str(it_lang.id)},
                }
            }
        }
    }
    record.commit()


def test_relations_field_pk_nested_list_of_objects_w_related_field(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with a list of objects."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']

    class Record1(Record, SystemFieldsMixin):
        # end result should be
        # {
        #   nested_array: [
        #       {languages: [en, fr]},
        #       {languages: [en, es]}
        #   ]
        # }
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key='nested_array_of_objects',
                keys=['iso'],
                record_cls=Language,
                relation_field='languages'
            ),
        )

    def _test_iterators(record):
        res_iter = record.relations.nested_array_of_objects()
        assert isinstance(res_iter, Iterable)
        # first inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == en_lang
        res = next(res_inner_iter)
        assert res == es_lang
        with pytest.raises(StopIteration):  # finished first iterator
            res = next(res_inner_iter)
        # second inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == fr_lang
        res = next(res_inner_iter)
        assert res == en_lang
        with pytest.raises(StopIteration):  # finished second iterator
            res = next(res_inner_iter)

    def _test_inner_values(record):
        with pytest.raises(KeyError):
            record['nested_array_of_objects'][0]['languages'][0]['id']
        assert record['nested_array_of_objects'][1]['languages'][0]['id'] == (
            str(en_lang.id))
        assert record['nested_array_of_objects'][1]['languages'][1]['id'] == (
            str(es_lang.id))
        assert record['nested_array_of_objects'][2]['languages'][0]['id'] == (
            str(fr_lang.id))
        assert record['nested_array_of_objects'][2]['languages'][1]['id'] == (
            str(en_lang.id))

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(
        Record1.relations.nested_array_of_objects, PKNestedListRelation
    )

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.nested_array_of_objects)
    assert record.relations.nested_array_of_objects() is None

    # Initialize from dictionary data
    record = Record1({'nested_array_of_objects': [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            }
        },
        {
            'languages': [
                {"id": str(en_lang.id)}, {"id": str(es_lang.id)}
            ],
            'field1': "string"
        },
        {
            'languages': [
                {"id": str(fr_lang.id)}, {"id": str(en_lang.id)}
            ],
            'field2': "string2"
        }
    ]})

    _test_iterators(record)

    # Initialize from field data
    record = Record1(
        {"metadata": {"title": "random title"}},
        relations={'nested_array_of_objects': [
            {
                'field1': "no language",
                'field2': "random string",
                'field3': {
                    'nestedField': 'nested string'
                }
            },
            {
                'languages': [str(en_lang.id), str(es_lang.id)],
                'field1': "string"
            },
            {
                'languages': [str(fr_lang.id), str(en_lang.id)],
                'field2': "string2"
            }
        ]})
    _test_inner_values(record)
    _test_iterators(record)

    # Set via record dictionary data
    record = Record1.create({})
    record['nested_array_of_objects'] = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            }
        },
        {
            'languages': [
                {"id": str(en_lang.id)}, {"id": str(es_lang.id)}
            ],
            'field1': "string"
        },
        {
            'languages': [
                {"id": str(fr_lang.id)}, {"id": str(en_lang.id)}
            ],
            'field2': "string2"
        }
    ]

    _test_iterators(record)

    record.commit()  # validates

    _test_iterators(record)

    # fails with non-list types
    record['nested_array_of_objects'] = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            },
            'languages': {'id': str(es_lang.id)}
        },
        {'field1': "testString"}
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # fails with invalid values
    record['nested_array_of_objects'] = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            },
            'languages': [{'id': 'invalid'}]
        },
        {'field1': "testString"}
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.nested_array_of_objects = [
        {
            'field1': "no language",
            'field2': "random string",
            'field3': {
                'nestedField': 'nested string'
            }
        },
        {
            'languages': [str(en_lang.id), str(es_lang.id)],
            'field1': "string"
        },
        {
            'languages': [str(fr_lang.id), str(en_lang.id)],
            'field2': "string2"
        }
    ]
    _test_inner_values(record)
    _test_iterators(record)

    # Update existing value
    record.relations.nested_array_of_objects = [
        {
            'languages': [str(fr_lang.id), str(en_lang.id)],
            'field2': "string2"
        }
    ]
    assert record['nested_array_of_objects'][0]['languages'][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_array_of_objects'][0]['languages'][1]['id'] == (
        str(en_lang.id))
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.nested_array_of_objects = 'invalid'
    # Check that old value is still there
    assert record['nested_array_of_objects'][0]['languages'][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_array_of_objects'][0]['languages'][1]['id'] == (
        str(en_lang.id))
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Dereference relation
    record.relations.nested_array_of_objects.dereference()
    assert record['nested_array_of_objects'] == [
        {
            'languages': [  # last updated values are fr + en
                {
                    'id': str(fr_lang.id),
                    'iso': 'fr',
                    '@v': str(fr_lang.id) + '::' + str(fr_lang.revision_id),
                }, {
                    'id': str(en_lang.id),
                    'iso': 'en',
                    '@v': str(en_lang.id) + '::' + str(en_lang.revision_id),
                },
            ],
            'field2': "string2"
        },
    ]

    # Commit clean dereferenced fields
    record.commit()
    assert record['nested_array_of_objects'][0]['languages'][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_array_of_objects'][0]['languages'][1]['id'] == (
        str(en_lang.id))

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record['nested_array_of_objects'][0]['languages'][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_array_of_objects'][0]['languages'][1]['id'] == (
        str(en_lang.id))
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Clear field
    record.relations.nested_array_of_objects = None
    assert 'nested_array_of_objects' not in record
    assert record.relations.nested_array_of_objects() is None
    record.commit()


def test_relations_field_pk_nested_list_of_objects_wo_related_field(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with a list of objects."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']

    class Record1(Record, SystemFieldsMixin):
        # end result should be
        # {
        #   nested_languages: [
        #       [en, fr],
        #       [es, fr],
        #   ]
        # }
        relations = RelationsField(
            nested_languages=PKNestedListRelation(
                key='nested_languages',
                keys=['iso'],
                record_cls=Language,
            ),
        )

    def _test_iterators(record):
        res_iter = record.relations.nested_languages()
        assert isinstance(res_iter, Iterable)
        # first inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == en_lang
        res = next(res_inner_iter)
        assert res == es_lang
        with pytest.raises(StopIteration):  # finished first iterator
            res = next(res_inner_iter)
        # second inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == fr_lang
        res = next(res_inner_iter)
        assert res == en_lang
        with pytest.raises(StopIteration):  # finished second iterator
            res = next(res_inner_iter)

    def _test_inner_values(record):
        assert record['nested_languages'][0][0]['id'] == (
            str(en_lang.id))
        assert record['nested_languages'][0][1]['id'] == (
            str(es_lang.id))
        assert record['nested_languages'][1][0]['id'] == (
            str(fr_lang.id))
        assert record['nested_languages'][1][1]['id'] == (
            str(en_lang.id))

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(
        Record1.relations.nested_languages, PKNestedListRelation
    )

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.nested_languages)
    assert record.relations.nested_languages() is None

    # Initialize from dictionary data
    record = Record1({'nested_languages': [
        [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
        [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}]
    ]})

    _test_iterators(record)

    # Initialize from field data
    record = Record1(
        {"metadata": {"title": "random title"}},
        relations={'nested_languages': [
            [str(en_lang.id), str(es_lang.id)],
            [str(fr_lang.id), str(en_lang.id)],
        ]})
    _test_inner_values(record)
    _test_iterators(record)

    # Set via record dictionary data
    record = Record1.create({})
    record['nested_languages'] = [
        [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
        [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
    ]

    _test_iterators(record)

    record.commit()  # validates

    _test_iterators(record)

    # fails with non-list types
    record['nested_languages'] = [
        {'id': str(es_lang.id)}
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # fails with invalid values
    record['nested_languages'] = [
        [{'id': 'invalid'}]
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.nested_languages = [
        [str(en_lang.id), str(es_lang.id)],
        [str(fr_lang.id), str(en_lang.id)],
    ]
    _test_inner_values(record)
    _test_iterators(record)

    # Update existing value
    record.relations.nested_languages = [
        [str(fr_lang.id), str(en_lang.id)],
    ]
    assert record['nested_languages'][0][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_languages'][0][1]['id'] == (
        str(en_lang.id))
    res_iter = record.relations.nested_languages()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.nested_languages = 'invalid'
    # Check that old value is still there
    assert record['nested_languages'][0][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_languages'][0][1]['id'] == (
        str(en_lang.id))
    res_iter = record.relations.nested_languages()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Dereference relation
    record.relations.nested_languages.dereference()
    assert record['nested_languages'] == [
        [  # last updated values are fr + en
            {
                'id': str(fr_lang.id),
                'iso': 'fr',
                '@v': str(fr_lang.id) + '::' + str(fr_lang.revision_id),
            }, {
                'id': str(en_lang.id),
                'iso': 'en',
                '@v': str(en_lang.id) + '::' + str(en_lang.revision_id),
            },
        ],
    ]

    # Commit clean dereferenced fields
    record.commit()
    assert record['nested_languages'][0][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_languages'][0][1]['id'] == (
        str(en_lang.id))

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record['nested_languages'][0][0]['id'] == (
        str(fr_lang.id))
    assert record['nested_languages'][0][1]['id'] == (
        str(en_lang.id))
    res_iter = record.relations.nested_languages()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Clear field
    record.relations.nested_languages = None
    assert 'nested_languages' not in record
    assert record.relations.nested_languages() is None
    record.commit()


def test_relations_field_pk_relation_with_value_check(testapp, db, languages):
    """RelationsField tests for PKRelation with value_check param."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']
    oe_lang = languages['oe']

    # Incorrect record definition with 1 non-list value to check
    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key='language',
                keys=['iso'],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity='English')),
            )
        )

    # Correct record definition with a list of values to check for
    class Record2(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key='language',
                keys=['iso'],
                record_cls=Language,
                value_check=dict(
                    information=dict(ethnicity=['English', 'French'])
                ),
            )
        )

    # Wrong record definition, value_check must contain existing fields in the
    # vocabulary value
    class Record3(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key='language',
                keys=['iso'],
                record_cls=Language,
                value_check=dict(wrong_value=dict(ethnicity=['English'])),
            )
        )

    # Correct record definition with multiple values to check
    class Record4(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key='language',
                keys=['iso'],
                record_cls=Language,
                value_check=dict(
                    information=dict(ethnicity=['English']), iso=['oe']
                ),
            )
        )

    record = Record1.create({})
    record['language'] = {'id': str(en_lang.id)}
    # Fails because value_check must be a list
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record2.create({})
    record['language'] = {'id': str(es_lang.id)}
    # Fails because only ethnicity accepted is English and French
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record['language'] = {'id': str(fr_lang.id)}
    record.commit()  # validates
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    record = Record3.create({})
    record['language'] = {'id': str(en_lang.id)}
    # Fails because wrong_value is not a field of the language vocabulary
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record4.create({})
    record['language'] = {'id': str(en_lang.id)}
    # Fails because only records with English ethnicity and 'oe' as iso are
    # accepted
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record['language'] = {'id': str(oe_lang.id)}
    record.commit()  # validates
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == oe_lang


def test_relations_field_pk_list_relation_with_value_check(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with value_check param."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']
    oe_lang = languages['oe']

    # Incorrect record definition with 1 non-list value to check
    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key='languages',
                keys=['iso', 'information.ethnicity'],
                record_cls=Language,
                value_check=dict(
                    information=dict(ethnicity='English')
                ),
            )
        )

    # Correct record definition with a list of values to check for
    class Record2(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key='languages',
                keys=['iso', 'information.ethnicity'],
                record_cls=Language,
                value_check=dict(
                    information=dict(ethnicity=['English', 'French'])
                ),
            )
        )

    # Wrong record definition, value_check must contain existing fields in the
    # vocabulary value
    class Record3(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key='languages',
                keys=['iso', 'information.ethnicity'],
                record_cls=Language,
                value_check=dict(wrong_value=dict(ethnicity=['English'])),
            )
        )

    # Correct record definition with multiple values to check
    class Record4(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key='languages',
                keys=['iso', 'information.ethnicity'],
                record_cls=Language,
                value_check=dict(
                    information=dict(ethnicity=['English']), iso=['oe']
                ),
            )
        )

    record = Record1.create({})
    record['languages'] = [{'id': str(fr_lang.id)}]
    # Fails because value_check must be a list
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record2.create({})
    record['languages'] = [{'id': str(es_lang.id)}, {'id': str(en_lang.id)}]
    # Fails because only ethnicity accepted is English and French
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record['languages'] = [{'id': str(fr_lang.id)}, {'id': str(en_lang.id)}]
    record.commit()  # validates
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    record = Record3.create({})
    record['languages'] = [{'id': str(en_lang.id)}]
    # Fails because wrong_value is not a field of the language vocabulary
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record4.create({})
    record['languages'] = [{'id': str(en_lang.id)}]
    # Fails because only records with English ethnicity and 'oe' as iso are
    # accepted
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record['languages'] = [{'id': str(oe_lang.id)}]
    record.commit()  # validates
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == oe_lang


def test_relations_field_pk_nested_list_of_obj_w_related_field_w_value_check(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with a list of objects and
    with the value_check param."""

    Language, languages = languages
    en_lang = languages['en']
    fr_lang = languages['fr']
    es_lang = languages['es']
    oe_lang = languages['oe']

    # Incorrect record definition with 1 non-list value to check
    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key='nested_array_of_objects',
                keys=['iso'],
                record_cls=Language,
                relation_field='languages',
                value_check=dict(
                    information=dict(ethnicity='English')
                ),
            ),
        )

    # Correct record definition with a list of values to check for
    class Record2(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key='nested_array_of_objects',
                keys=['iso'],
                record_cls=Language,
                relation_field='languages',
                value_check=dict(
                    information=dict(ethnicity=['English', 'French'])
                ),
            ),
        )

    # Wrong record definition, value_check must contain existing fields in the
    # vocabulary value
    class Record3(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key='nested_array_of_objects',
                keys=['iso'],
                record_cls=Language,
                relation_field='languages',
                value_check=dict(wrong_value=dict(ethnicity=['English'])),
            ),
        )

    # Correct record definition with multiple values to check
    class Record4(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key='nested_array_of_objects',
                keys=['iso'],
                record_cls=Language,
                relation_field='languages',
                value_check=dict(
                    information=dict(ethnicity=['English']), iso=['oe']
                ),
            ),
        )

    record = Record1.create({})
    record['nested_array_of_objects'] = [
        {
            'languages': [
                {"id": str(en_lang.id)}, {"id": str(es_lang.id)}
            ],
            'field1': "string"
        },
        {
            'languages': [
                {"id": str(fr_lang.id)}, {"id": str(en_lang.id)}
            ],
            'field2': "string2"
        }
    ]
    # Fails because value_check must be a list
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record2.create({})
    record['nested_array_of_objects'] = [
        {
            'languages': [
                {"id": str(en_lang.id)}, {"id": str(es_lang.id)}
            ],
            'field1': "string"
        },
        {
            'languages': [
                {"id": str(fr_lang.id)}, {"id": str(en_lang.id)}
            ],
            'field2': "string2"
        }
    ]
    # Fails because only ethnicity accepted is English and French
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record['nested_array_of_objects'] = [
        {
            'languages': [
                {"id": str(en_lang.id)}, {"id": str(fr_lang.id)}
            ],
            'field1': "string"
        },
        {
            'languages': [
                {"id": str(fr_lang.id)}, {"id": str(en_lang.id)}
            ],
            'field2': "string2"
        }
    ]

    record.commit()  # validates
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == en_lang
    res = next(res_inner_iter)
    assert res == fr_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)
    # second inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished second iterator
        res = next(res_inner_iter)

    record = Record3.create({})
    record['nested_array_of_objects'] = [
        {
            'languages': [
                {"id": str(en_lang.id)}, {"id": str(fr_lang.id)}
            ],
            'field1': "string"
        },
        {
            'languages': [
                {"id": str(fr_lang.id)}, {"id": str(en_lang.id)}
            ],
            'field2': "string2"
        }
    ]
    # Fails because wrong_value is not a field of the language vocabulary
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record4.create({})
    record['nested_array_of_objects'] = [
        {
            'languages': [
                {"id": str(en_lang.id)}
            ],
            'field1': "string"
        }
    ]
    # Fails because only ethnicity accepted is English
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record['nested_array_of_objects'] = [
        {
            'languages': [
                {"id": str(oe_lang.id)}
            ],
            'field1': "string"
        }
    ]
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == oe_lang
    with pytest.raises(StopIteration):  # finished first iterator
        next(res_inner_iter)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test admin interface."""


import uuid
from unittest.mock import patch

from flask import url_for
from flask_admin import Admin, menu
from sqlalchemy.exc import SQLAlchemyError

from invenio_records.admin import record_adminview
from invenio_records.api import Record


def test_admin(testapp, db):
    """Test flask-admin interace."""
    admin = Admin(testapp, name="Test")

    assert 'model' in record_adminview
    assert 'modelview' in record_adminview

    # Register both models in admin
    model = record_adminview.pop('model')
    view = record_adminview.pop('modelview')
    admin.add_view(view(model, db.session, **record_adminview))

    # Check if generated admin menu contains the correct items
    menu_items = {str(item.name): item for item in admin.menu()}
    assert 'Records' in menu_items
    assert menu_items['Records'].is_category()

    submenu_items = {
        str(item.name): item for item in menu_items['Records'].get_children()}
    assert 'Record Metadata' in submenu_items
    assert isinstance(submenu_items['Record Metadata'], menu.MenuView)

    # Create a test record.
    rec_uuid = str(uuid.uuid4())
    Record.create({'title': 'test<script>alert(1);</script>'}, id_=rec_uuid)
    db.session.commit()

    with testapp.test_request_context():
        index_view_url = url_for('recordmetadata.index_view')
        delete_view_url = url_for('recordmetadata.delete_view')
        detail_view_url = url_for(
            'recordmetadata.details_view', id=rec_uuid)

    with testapp.test_client() as client:
        # List index view and check record is there.
        res = client.get(index_view_url)
        assert res.status_code == 200

        # Check for XSS in JSON output
        res = client.get(detail_view_url)
        assert res.status_code == 200
        data = res.get_data(as_text=True)
        assert '<pre>{' in data
        assert '}</pre>' in data
        assert '<script>alert(1);</script>' not in data

        # Fake a problem with SQLAlchemy.
        with patch('invenio_records.models.RecordMetadata') as db_mock:
            db_mock.side_effect = SQLAlchemyError()
            res = client.post(
                delete_view_url, data={'id': rec_uuid}, follow_redirects=True)
            assert res.status_code == 200

        # Delete it.
        res = client.post(
            delete_view_url, data={'id': rec_uuid}, follow_redirects=True)
        assert res.status_code == 200

        # View the delete record
        res = client.get(detail_view_url)
        assert res.status_code == 200
        assert '<pre>null</pre>' in res.get_data(as_text=True)

        # Delete it again
        res = client.post(
            delete_view_url, data={'id': rec_uuid}, follow_redirects=True)
        assert res.status_code == 200



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field."""

from ..base import SystemField
from .mapping import RelationsMapping
from .relations import RelationBase


#
# System field
#
class RelationsField(SystemField):
    """Relations field for connections to external entities."""

    def __init__(self, **fields):
        """Initialize the field."""
        assert all(isinstance(f, RelationBase) for f in fields.values())
        self._fields = fields

    def __getattr__(self, name):
        """Get a field definition."""
        if name in self._fields:
            return self._fields[name]
        raise AttributeError

    def __iter__(self):
        """Iterate over the configured fields."""
        return iter(getattr(self, f) for f in self._fields)

    def __contains__(self, name):
        """Return if a field exists in the configured fields."""
        return name in self._fields

    #
    # Helpers
    #
    def obj(self, instance):
        """Get the relations object."""
        # Check cache
        obj = self._get_cache(instance)
        if obj:
            return obj
        obj = RelationsMapping(record=instance, fields=self._fields)
        self._set_cache(instance, obj)
        return obj

    #
    # Data descriptor
    #
    def __get__(self, record, owner=None):
        """Accessing the attribute."""
        # Class access
        if record is None:
            return self
        return self.obj(record)

    def __set__(self, instance, values):
        """Setting the attribute."""
        obj = self.obj(instance)
        for k, v in values.items():
            setattr(obj, k, v)

    #
    # Record extension
    #
    def pre_commit(self, record):
        """Initialise the model field."""
        self.obj(record).validate()
        self.obj(record).clean()
